"""
Vault (or OpenBao) API client implementation
"""

import logging
import random
import re
from itertools import takewhile

import requests
import salt.exceptions
from requests.adapters import HTTPAdapter
from requests.adapters import Retry

from saltext.vault.utils.vault import leases
from saltext.vault.utils.vault.exceptions import VaultAuthExpired
from saltext.vault.utils.vault.exceptions import VaultInvocationError
from saltext.vault.utils.vault.exceptions import VaultNotFoundError
from saltext.vault.utils.vault.exceptions import VaultPermissionDeniedError
from saltext.vault.utils.vault.exceptions import VaultPreconditionFailedError
from saltext.vault.utils.vault.exceptions import VaultRateLimitExceededError
from saltext.vault.utils.vault.exceptions import VaultServerError
from saltext.vault.utils.vault.exceptions import VaultUnavailableError
from saltext.vault.utils.vault.exceptions import VaultUnsupportedOperationError
from saltext.vault.utils.vault.exceptions import VaultUnwrapException

try:
    from urllib3.util import create_urllib3_context

    URLLIB3V1 = False
except ImportError:
    # urllib <2
    from urllib3.util.ssl_ import create_urllib3_context

    URLLIB3V1 = True


log = logging.getLogger(__name__)
logging.getLogger("requests").setLevel(logging.WARNING)

# This list is not complete at all, but contains
# the most important paths.
VAULT_UNAUTHD_PATHS = (
    "sys/wrapping/lookup",
    "sys/internal/ui/mounts",
    "sys/internal/ui/namespaces",
    "sys/seal-status",
    "sys/health",
)

HTTP_TOO_MANY_REQUESTS = 429

# Default timeout configuration
DEFAULT_CONNECT_TIMEOUT = 9.2
DEFAULT_READ_TIMEOUT = 30

# Default retry configuration
DEFAULT_MAX_RETRIES = 5
DEFAULT_BACKOFF_FACTOR = 0.1
DEFAULT_BACKOFF_MAX = 10.0
DEFAULT_BACKOFF_JITTER = 0.2
DEFAULT_RETRY_POST = False
DEFAULT_RESPECT_RETRY_AFTER = True
DEFAULT_RETRY_AFTER_MAX = 60
# https://developer.hashicorp.com/vault/api-docs#http-status-codes
# 412: eventually consistent data is still missing (Enterprise)
DEFAULT_RETRY_STATUS = (412, 500, 502, 503, 504)

# Caps for retry configuration
MAX_MAX_RETRIES = 10
MAX_BACKOFF_FACTOR = 3.0
MAX_BACKOFF_MAX = 60.0
MAX_BACKOFF_JITTER = 5.0


def _get_expected_creation_path(secret_type, config=None):
    if secret_type == "token":
        return r"auth/token/create(/[^/]+)?"

    if secret_type == "secret_id":
        if config is not None:
            mount, approle = re.escape(config["auth"]["approle_mount"]), re.escape(
                config["auth"]["approle_name"]
            )
            return rf"auth/{mount}/role/{approle}/secret\-id"
        return r"auth/[^/]+/role/[^/]+/secret\-id"

    if secret_type == "role_id":
        if config is not None:
            mount, approle = re.escape(config["auth"]["approle_mount"]), re.escape(
                config["auth"]["approle_name"]
            )
            return rf"auth/{mount}/role/{approle}/role\-id"
        return r"auth/[^/]+/role/[^/]+/role\-id"

    raise salt.exceptions.SaltInvocationError(
        f"secret_type must be one of token, secret_id, role_id, got `{secret_type}`."
    )


class VaultClient:
    """
    Unauthenticated client for the Vault API.
    Base class for authenticated client.
    """

    def __init__(
        self,
        url,
        namespace=None,
        verify=None,
        session=None,
        connect_timeout=DEFAULT_CONNECT_TIMEOUT,
        read_timeout=DEFAULT_READ_TIMEOUT,
        max_retries=DEFAULT_MAX_RETRIES,
        backoff_factor=DEFAULT_BACKOFF_FACTOR,
        backoff_max=DEFAULT_BACKOFF_MAX,
        backoff_jitter=DEFAULT_BACKOFF_JITTER,
        retry_post=DEFAULT_RETRY_POST,
        respect_retry_after=DEFAULT_RESPECT_RETRY_AFTER,
        retry_status=DEFAULT_RETRY_STATUS,
        retry_after_max=DEFAULT_RETRY_AFTER_MAX,
    ):
        self.url = url
        self.namespace = namespace
        self.verify = verify

        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout

        # Cap the retry-backoff values somewhat
        self.max_retries = max(0, min(max_retries, MAX_MAX_RETRIES))
        self.backoff_factor = max(0, min(backoff_factor, MAX_BACKOFF_FACTOR))
        self.backoff_max = max(0, min(backoff_max, MAX_BACKOFF_MAX))
        self.backoff_jitter = max(0, min(backoff_jitter, MAX_BACKOFF_JITTER))
        self.retry_post = bool(retry_post)
        self.respect_retry_after = bool(respect_retry_after)
        # urllib3 2.6.3 introduced this parameter and set its default to 21600 (6h). Match that.
        self.retry_after_max = max(0, retry_after_max) if retry_after_max is not None else 21600
        self.retry_status = tuple(retry_status) if retry_status is not None else None

        retry = VaultRetry(
            total=self.max_retries,
            backoff_factor=self.backoff_factor,
            backoff_max=self.backoff_max,
            backoff_jitter=self.backoff_jitter,
            respect_retry_after_header=self.respect_retry_after,
            retry_after_max=self.retry_after_max,
            allowed_methods=None if retry_post else Retry.DEFAULT_ALLOWED_METHODS,
            raise_on_status=False,
            status_forcelist=self.retry_status,
        )

        if session is None:
            session = requests.Session()
            adapter = VaultAPIAdapter(
                max_retries=retry,
                verify=verify,
                connect_timeout=self.connect_timeout,
                read_timeout=self.read_timeout,
            )
            session.mount(url, adapter)
        else:
            # Sessions should only be inherited from other instances
            # of this class. A changed ``verify`` setting causes a fresh
            # client to be instantiated.
            # We want to keep the TCP connection alive, so we'll modify
            # the adapter in place.
            adapter = session.get_adapter(url)
            adapter.max_retries = retry
            adapter.connect_timeout = self.connect_timeout
            adapter.read_timeout = self.read_timeout
        self.session = session
        self._vault_adapter = adapter

    def delete(
        self, endpoint, *, wrap=False, raise_error=True, add_headers=None, safe_to_retry=None
    ):
        """
        Wrapper for client.request("DELETE", ...)
        """
        return self.request(
            "DELETE",
            endpoint,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
        )

    def get(self, endpoint, *, wrap=False, raise_error=True, add_headers=None, safe_to_retry=None):
        """
        Wrapper for client.request("GET", ...)
        """
        return self.request(
            "GET",
            endpoint,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
        )

    def list(self, endpoint, *, wrap=False, raise_error=True, add_headers=None, safe_to_retry=None):
        """
        Wrapper for client.request("LIST", ...)
        TODO: configuration to enable GET requests with query parameters for LIST?
        """
        return self.request(
            "LIST",
            endpoint,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
        )

    def post(
        self,
        endpoint,
        payload=None,
        *,
        wrap=False,
        raise_error=True,
        add_headers=None,
        safe_to_retry=None,
    ):
        """
        Wrapper for client.request("POST", ...)
        """
        return self.request(
            "POST",
            endpoint,
            payload=payload,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
        )

    def put(
        self,
        endpoint,
        payload=None,
        *,
        wrap=False,
        raise_error=True,
        add_headers=None,
        safe_to_retry=True,
    ):
        """
        Also a wrapper for client.request("POST", ...)
        Vault considers POST and PUT to be synonymous.
        The difference to ``post`` is that this request is marked as safe to retry by default (idempotent).
        """
        return self.request(
            "POST",
            endpoint,
            payload=payload,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
        )

    def patch(
        self,
        endpoint,
        payload,
        *,
        wrap=False,
        raise_error=True,
        add_headers=None,
        safe_to_retry=None,
    ):
        """
        Wrapper for client.request("PATCH", ...)
        """
        return self.request(
            "PATCH",
            endpoint,
            payload=payload,
            wrap=wrap,
            raise_error=raise_error,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
        )

    def request(
        self,
        method,
        endpoint,
        payload=None,
        *,
        wrap=False,
        raise_error=True,
        add_headers=None,
        safe_to_retry=None,
        **kwargs,
    ):
        """
        Issue a request against the Vault API.
        Returns boolean when no data was returned, otherwise the decoded json data
        or a VaultWrappedResponse object if wrapping was requested.
        """
        res = self.request_raw(
            method,
            endpoint,
            payload=payload,
            wrap=wrap,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
            **kwargs,
        )
        if res.status_code == 204:
            return True
        data = res.json()
        if not res.ok:
            if raise_error:
                self._raise_status(res)
            return data
        if wrap:
            return leases.VaultWrappedResponse(**data["wrap_info"])
        return data

    def request_raw(
        self,
        method,
        endpoint,
        payload=None,
        *,
        wrap=False,
        add_headers=None,
        safe_to_retry=None,
        **kwargs,
    ):
        """
        Issue a request against the Vault API. Returns the raw response object.
        """
        url = self._get_url(endpoint)
        headers = self._get_headers(wrap)
        if method.upper() == "PATCH":
            # PATCH always requires JSON patch content-type, so
            # just replace it.
            headers["Content-Type"] = "application/merge-patch+json"

        try:
            headers.update(add_headers)
        except TypeError:
            pass

        self._vault_adapter.max_retries.safe_to_retry = safe_to_retry

        try:
            res = self.session.request(
                method,
                url,
                headers=headers,
                json=payload,
                **kwargs,
            )
        finally:
            self._vault_adapter.max_retries.safe_to_retry = None

        return res

    def unwrap(self, wrapped, expected_creation_path=None):
        """
        Unwraps the data associated with a wrapping token.

        wrapped
            Wrapping token to unwrap

        expected_creation_path
            Regex expression or list of expressions that should fully match the
            wrapping token creation path. At least one match is required.
            Defaults to None, which skips the check.

            .. note::
                This check prevents tampering with wrapping tokens, which are
                valid for one request only. Usually, if an attacker sniffs a wrapping
                token, there will be two unwrapping requests, causing an audit warning.
                If the attacker can issue a new wrapping token and insert it into the
                response instead, this warning would be silenced. Assuming they do not
                possess the permissions to issue a wrapping token from the correct
                endpoint, checking the creation path makes this kind of attack obvious.
        """
        if expected_creation_path:
            wrap_info = self.wrap_info(wrapped)
            if not isinstance(expected_creation_path, list):
                expected_creation_path = [expected_creation_path]
            if not any(re.fullmatch(p, wrap_info["creation_path"]) for p in expected_creation_path):
                raise VaultUnwrapException(
                    actual=wrap_info["creation_path"],
                    expected=expected_creation_path,
                    url=self.url,
                    namespace=self.namespace,
                    verify=self.verify,
                )
        url = self._get_url("sys/wrapping/unwrap")
        headers = self._get_headers()
        payload = {}
        if "X-Vault-Token" not in headers:
            headers["X-Vault-Token"] = str(wrapped)
        else:
            payload["token"] = str(wrapped)
        res = self.session.request("POST", url, headers=headers, json=payload)
        if not res.ok:
            self._raise_status(res)
        return res.json()

    def wrap_info(self, wrapped):
        """
        Lookup wrapping token meta information.
        """
        endpoint = "sys/wrapping/lookup"
        add_headers = {"X-Vault-Token": str(wrapped)}
        return self.put(endpoint, wrap=False, add_headers=add_headers)["data"]

    def token_lookup(self, token=None, accessor=None, raw=False):
        """
        Lookup token meta information.

        token
            The token to look up or to use to look up the accessor.
            Required.

        accessor
            The accessor to use to query the token meta information.

        raw
            Return the raw response object instead of response data.
            Also disables status code checking.
        """
        endpoint = "auth/token/lookup-self"
        method = "GET"
        payload = {}
        if token is None:
            raise VaultInvocationError("Unauthenticated VaultClient needs a token to lookup.")
        add_headers = {"X-Vault-Token": token}

        if accessor is not None:
            method = "POST"
            endpoint = "auth/token/lookup-accessor"
            payload["accessor"] = accessor

        res = self.request_raw(
            method,
            endpoint,
            payload=payload,
            wrap=False,
            add_headers=add_headers,
            safe_to_retry=True,
        )
        if raw:
            return res
        self._raise_status(res)
        return res.json()["data"]

    def token_valid(self, valid_for=0, remote=True):  # pylint: disable=unused-argument
        """
        This client does not have a token, hence it's always invalid.
        """
        return False

    def get_config(self):
        """
        Returns Vault server configuration used by this client.
        """
        return {
            "url": self.url,
            "namespace": self.namespace,
            "verify": self.verify,
        }

    def _get_url(self, endpoint):
        endpoint = endpoint.strip("/")
        return f"{self.url}/v1/{endpoint}"

    def _get_headers(self, wrap=False):
        headers = {"Content-Type": "application/json", "X-Vault-Request": "true"}
        if self.namespace is not None:
            headers["X-Vault-Namespace"] = self.namespace
        if wrap:
            headers["X-Vault-Wrap-TTL"] = str(wrap)
        return headers

    def _raise_status(self, res):
        errors = ", ".join(res.json().get("errors", []))
        if res.status_code == 400:
            raise VaultInvocationError(errors)
        if res.status_code == 403:
            raise VaultPermissionDeniedError(errors)
        if res.status_code == 404:
            raise VaultNotFoundError(errors)
        if res.status_code == 405:
            raise VaultUnsupportedOperationError(errors)
        if res.status_code == 412:
            raise VaultPreconditionFailedError(errors)
        if res.status_code == HTTP_TOO_MANY_REQUESTS:
            raise VaultRateLimitExceededError(errors)
        if res.status_code in (500, 502):
            raise VaultServerError(errors)
        if res.status_code == 503:
            raise VaultUnavailableError(errors)
        res.raise_for_status()


class AuthenticatedVaultClient(VaultClient):
    """
    Authenticated client for the Vault API.
    This should be used for most operations.
    """

    auth = None

    def __init__(self, auth, url, **kwargs):
        self.auth = auth
        super().__init__(url, **kwargs)

    def token_valid(self, valid_for=0, remote=True):
        """
        Check whether this client's authentication information is
        still valid.

        remote
            Check with the remote Vault server as well. This consumes
            a token use. Defaults to true.
        """
        if not self.auth.is_valid(valid_for):
            return False
        if not remote:
            return True
        try:
            res = self.token_lookup(raw=True)
            if res.status_code != 200:
                return False
            return True
        except Exception as err:  # pylint: disable=broad-except
            raise salt.exceptions.CommandExecutionError(
                "Error while looking up self token."
            ) from err

    def token_lookup(self, token=None, accessor=None, raw=False):
        """
        Lookup token meta information.

        token
            The token to look up. If neither token nor accessor
            are specified, looks up the current token in use by
            this client.

        accessor
            The accessor of the token to query the meta information for.

        raw
            Return the raw response object instead of response data.
            Also disables status code checking.
        """
        endpoint = "auth/token/lookup"
        method = "POST"
        payload = {}
        if token is None and accessor is None:
            endpoint += "-self"
            method = "GET"
        if token is not None:
            payload["token"] = token
        elif accessor is not None:
            endpoint += "-accessor"
            payload["accessor"] = accessor
        if raw:
            return self.request_raw(
                method, endpoint, payload=payload, wrap=False, safe_to_retry=True
            )
        return self.request(method, endpoint, payload=payload, wrap=False, safe_to_retry=True)[
            "data"
        ]

    def token_renew(self, increment=None, token=None, accessor=None):
        """
        Renew a token.

        increment
            Request the token to be valid for this amount of time from the current
            point of time onwards. Can also be used to reduce the validity period.
            The server might not honor this increment.
            Can be an integer (seconds) or a time string like ``1h``. Optional.

        token
            The token that should be renewed. Optional.
            If token and accessor are unset, renews the token currently in use
            by this client.

        accessor
            The accessor of the token that should be renewed. Optional.
        """
        endpoint = "auth/token/renew"
        payload = {}

        if token is None and accessor is None:
            if not self.auth.is_renewable():
                return False
            endpoint += "-self"

        if increment is not None:
            payload["increment"] = increment
        if token is not None:
            payload["token"] = token
        elif accessor is not None:
            endpoint += "-accessor"
            payload["accessor"] = accessor

        res = self.post(endpoint, payload=payload)

        if token is None and accessor is None:
            self.auth.update_token(res["auth"])
        return res["auth"]

    def token_revoke(self, delta=1, token=None, accessor=None):
        """
        Revoke a token by setting its TTL to 1s.

        delta
            The time in seconds to request revocation after.
            Defaults to 1s.

        token
            The token that should be revoked. Optional.
            If token and accessor are unset, revokes the token currently in use
            by this client.

        accessor
            The accessor of the token that should be revoked. Optional.
        """
        try:
            self.token_renew(increment=delta, token=token, accessor=accessor)
        except (VaultPermissionDeniedError, VaultNotFoundError, VaultAuthExpired):
            # if we're trying to revoke ourselves and this happens,
            # the token was already invalid
            if token or accessor:
                raise
            return False
        return True

    def request_raw(
        self,
        method,
        endpoint,
        payload=None,
        *,
        wrap=False,
        add_headers=None,
        safe_to_retry=None,
        is_unauthd=False,
        **kwargs,
    ):  # pylint: disable=arguments-differ
        """
        Issue an authenticated request against the Vault API. Returns the raw response object.
        """
        ret = super().request_raw(
            method,
            endpoint,
            payload=payload,
            wrap=wrap,
            add_headers=add_headers,
            safe_to_retry=safe_to_retry,
            **kwargs,
        )
        # tokens are used regardless of status code
        if not is_unauthd and not endpoint.startswith(VAULT_UNAUTHD_PATHS):
            self.auth.used()
        return ret

    def _get_headers(self, wrap=False):
        headers = super()._get_headers(wrap)
        headers["X-Vault-Token"] = str(self.auth.get_token())
        return headers


class VaultAPIAdapter(HTTPAdapter):
    """
    An adapter that

        * allows to restrict requests CA chain validation to a single
          root certificate without writing it to disk.
        * sets default values for timeout settings without having to
          specify it in every request.
    """

    def __init__(self, *args, verify=None, connect_timeout=None, read_timeout=None, **kwargs):
        ca_cert_data = None
        try:
            if verify.strip().startswith("-----BEGIN CERTIFICATE"):
                ca_cert_data = verify
                verify = None
        except AttributeError:
            pass
        self.ca_cert_data = ca_cert_data
        self.verify = verify
        self.connect_timeout = connect_timeout or DEFAULT_CONNECT_TIMEOUT
        self.read_timeout = read_timeout or DEFAULT_READ_TIMEOUT
        super().__init__(*args, **kwargs)

    def init_poolmanager(
        self,
        connections,
        maxsize,
        block=requests.adapters.DEFAULT_POOLBLOCK,
        **pool_kwargs,
    ):
        if self.ca_cert_data is not None:
            ssl_context = create_urllib3_context()
            ssl_context.load_verify_locations(cadata=self.ca_cert_data)
            pool_kwargs["ssl_context"] = ssl_context
        return super().init_poolmanager(connections, maxsize, block=block, **pool_kwargs)

    def send(self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None):
        """
        Wrap sending the request to ensure ``verify`` and ``timeout`` is set
        as specified on every request. ``timeout`` can be overridden per request.
        """
        if self.verify is not None:
            verify = self.verify
        if timeout is None:
            timeout = (self.connect_timeout, self.read_timeout)
        return super().send(
            request, stream=stream, timeout=timeout, verify=verify, cert=cert, proxies=proxies
        )


class VaultRetry(Retry):
    """
    The Vault API responds with HTTP 429 when rate limits have been hit.
    We want to always retry 429, regardless of the HTTP verb and the presence
    of the ``Retry-After`` header, thus we need to subclass the retry configuration class.
    For HTTP error responses, we do not want to retry immediately if the header was not set.

    We override the default exponential power-of-2 algorithm for calculating
    the backoff time with a Fibonacci one because we expect a relatively
    quick turnaround.
    """

    PHI = 1.618
    SQRT5 = 2.236

    def __init__(
        self,
        *args,
        backoff_jitter=0.0,
        backoff_max=Retry.DEFAULT_BACKOFF_MAX,
        retry_after_max=DEFAULT_RETRY_AFTER_MAX,
        **kwargs,
    ):
        """
        For ``urllib3<2``, backport ``backoff_max`` and ``backoff_jitter``.
        Also, allow limiting the value returned by ``Retry-After`` by
        specifying ``retry_after_max``.
        """
        if URLLIB3V1:
            self.backoff_max = backoff_max
            self.backoff_jitter = backoff_jitter
        else:
            kwargs["backoff_max"] = backoff_max
            kwargs["backoff_jitter"] = backoff_jitter
        super().__init__(*args, **kwargs)
        # urllib3 2.6.3 introduced the same parameter. Avoid having to guess
        # wheter the parameter is supported by setting it ourselves.
        self.retry_after_max = retry_after_max
        self.safe_to_retry = None

    def _is_method_retryable(self, method):
        """
        Some API calls can be safely retried, even if their HTTP method is not
        considered idempotent usually (e.g. POST for lookups).
        Other HTTP methods are generally considered idempotent, but are not
        for specific API calls (like DELETE on KVv2).

        Therefore, we allow to override the inbuilt HTTP method-only logic
        on a per-request basis.
        """
        if self.safe_to_retry is not None:
            return self.safe_to_retry
        return super()._is_method_retryable(method)

    def is_retry(self, method, status_code, has_retry_after=False):
        """
        HTTP 429 is always retryable (even for POST/PATCH), otherwise fall back
        to the configuration.
        """
        if status_code == HTTP_TOO_MANY_REQUESTS:
            return True
        return super().is_retry(method, status_code, has_retry_after=has_retry_after)

    def get_backoff_time(self):
        """
        When we're retrying HTTP error responses, ensure we don't execute the
        first retry immediately.
        Also overrides the default 2**n algorithm with one based on the Fibonacci sequence.
        On ``urllib3<2``, this also backports ``backoff_jitter`` and ``backoff_max``.
        """
        # We want to consider only the last consecutive errors sequence (Ignore redirects).
        consecutive_errors = list(
            takewhile(lambda x: x.redirect_location is None, reversed(self.history))
        )
        consecutive_errors_len = len(consecutive_errors)
        if consecutive_errors_len and consecutive_errors[0].status is not None:
            # Ensure we only immediately retry for local (connection/read) errors,
            # not when we got an HTTP response.
            consecutive_errors_len += 1
        if consecutive_errors_len <= 1:
            return 0
        # Approximate the nth Fibonacci number.
        # We want to begin with the 4th one (2).
        backoff_value = round(
            self.backoff_factor * round(self.PHI ** (consecutive_errors_len + 1) / self.SQRT5),
            1,
        )
        if self.backoff_jitter != 0.0:
            backoff_value += random.random() * self.backoff_jitter
        return float(max(0, min(self.backoff_max, backoff_value)))

    def get_retry_after(self, response):
        """
        The default implementation sleeps for as long as requested
        by the ``Retry-After`` header. We want to limit that somewhat
        to avoid sleeping until the end of the universe.
        """
        retry_after = response.headers.get("Retry-After")

        if retry_after is None:
            return None

        res = self.parse_retry_after(retry_after)
        if self.retry_after_max is None:
            return res
        return min(res, self.retry_after_max)

    def new(self, **kw):
        """
        Since we backport some params and introduce a new one,
        ensure all requests use the defined parameters, not the default ones.
        """
        ret = super().new(**kw)
        if URLLIB3V1:
            ret.backoff_jitter = self.backoff_jitter
            ret.backoff_max = self.backoff_max
        ret.retry_after_max = self.retry_after_max
        ret.safe_to_retry = self.safe_to_retry
        return ret



"""
Use secrets sourced from Vault (or OpenBao) in minion pillars.

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.

.. warning::
    A minion must not be able to write to its own pillar source path,
    otherwise a core security assumption in Salt is violated.

.. versionchanged:: 1.0.0
    Previous versions of this pillar module found in Salt core were configured
    with a parameter named ``conf``, expecting a single value representing
    the path to include in the pillar with the prefix ``path=``.
    This parameter has been deprecated. Please configure this pillar module
    either by just passing the path or declaring it as ``path: <path>``.

Setup
-----
Include this module in your :conf_master:`ext_pillar` configuration:

.. code-block:: yaml

    ext_pillar:
      - vault: salt/global

.. hint::
    You can also include multiple instances of this module in your configuration.

Now all keys of the Vault KV path ``salt/global`` are inserted into each
minion's pillar, which is quite inflexible and usually not what is wanted.
To work around that, you can :ref:`template the path <vault-templating>`.

.. code-block:: yaml

    ext_pillar:
      - vault: salt/minions/{minion}
      - vault: salt/roles/{pillar[roles]}

.. note::
    There is currently no ``top.sls`` equivalent.

.. note::
    If a pattern matches multiple paths, the results are merged according
    to the master configuration values :conf_master:`pillar_source_merging_strategy <pillar_source_merging_strategy>`
    and :conf_master:`pillar_merge_lists <pillar_merge_lists>` by default.
    If the optional :vconf:`nesting_key <pillar.nesting_key>` is defined,
    the merged result is nested below.
    There is currently no way to nest multiple results under different keys.

.. vconf:: pillar

Configuration reference
-----------------------
.. vconf:: pillar.path

``path``
    The path to include in the minion pillars. Can be :ref:`templated <vault-templating>`.

.. vconf:: pillar.nesting_key

``nesting_key``
    The Vault-sourced pillar values are usually merged into the root
    of the pillar. This option allows you to specify a parent key
    under which all values are nested. If the key contains previous
    values, they are merged.

.. vconf:: pillar.merge_strategy

``merge_strategy``
    When multiple paths are matched by a templated path, use this merge strategy
    instead of :conf_master:`pillar_source_merging_strategy <pillar_source_merging_strategy>`.

.. vconf:: pillar.merge_lists

``merge_lists``
    Override the default set by :conf_master:`pillar_merge_lists <pillar_merge_lists>`.


Complete configuration
----------------------
.. code-block:: yaml

    ext_pillar:
      - vault:
           path: salt/roles/{pillar[roles]}
           nesting_key: vault_sourced
           merge_strategy: smart
           merge_lists: false
"""

import logging

import salt.utils.dictupdate
from salt.exceptions import InvalidConfigError
from salt.exceptions import SaltException

from saltext.vault.utils import vault
from saltext.vault.utils.vault import helpers
from saltext.vault.utils.versions import warn_until

log = logging.getLogger(__name__)


def ext_pillar(
    minion_id,  # pylint: disable=W0613
    pillar,  # pylint: disable=W0613
    path=None,
    nesting_key=None,
    merge_strategy=None,
    merge_lists=None,
    extra_minion_data=None,
    conf=None,
):
    """
    Get pillar data from Vault for the configuration ``conf``.
    """
    extra_minion_data = extra_minion_data or {}
    if extra_minion_data.get("_vault_runner_is_compiling_pillar_templates"):
        # Disable vault ext_pillar while compiling pillar for vault policy templates
        return {}
    if conf is not None:
        comps = conf.split()
        paths = [comp for comp in comps if comp.startswith("path=")]
        if not paths:
            log.error('"%s" is not a valid Vault ext_pillar config', conf)
            return {}
        path_pattern = paths[0].replace("path=", "")
        warn_until(
            2,
            (
                "The `conf` parameter to the Vault pillar is deprecated. "
                "Please migrate to the `path` parameter. It takes the path "
                "as its parameter, without the `path=` prefix."
            ),
        )
    elif path is not None:
        comps = path.split()
        if not comps:
            log.error('"%s" is not a valid Vault ext_pillar config', path)
            return {}
        if len(comps) > 1:
            warn_until(
                2,
                (
                    "The `conf` parameter to the Vault pillar is deprecated. "
                    "Please migrate to the `path` parameter. It takes the path "
                    "as its parameter, without the `path=` prefix."
                ),
            )
            comps = [comp for comp in comps if comp.startswith("path=")]
            if not comps:
                log.error('"%s" is not a valid Vault ext_pillar config', path)
                return {}
        path = comps[0]
        if path.startswith("path="):
            warn_until(
                2,
                (
                    "The Vault pillar module should not be configured with "
                    "the `path=` prefix anymore. Please remove it from your "
                    "configuration."
                ),
            )
            path = path[5:]
        path_pattern = path
    else:
        raise InvalidConfigError("Need a Vault path to include in the pillar")

    merge_strategy = merge_strategy or __opts__.get("pillar_source_merging_strategy", "smart")
    merge_lists = merge_lists or __opts__.get("pillar_merge_lists", False)

    vault_pillar = {}

    for rendered_path in _get_paths(path_pattern, minion_id, pillar):
        try:
            vault_pillar_single = vault.read_kv(rendered_path, __opts__, __context__)
        except vault.VaultNotFoundError:
            log.info(
                "Vault secret not found for: %s", rendered_path, exc_info_on_loglevel=logging.DEBUG
            )
        except SaltException:
            log.warning(
                "Error fetching Vault secret at: %s",
                rendered_path,
                exc_info_on_loglevel=logging.DEBUG,
            )
        else:
            vault_pillar = salt.utils.dictupdate.merge(
                vault_pillar,
                vault_pillar_single,
                strategy=merge_strategy,
                merge_lists=merge_lists,
            )

    if nesting_key:
        vault_pillar = {nesting_key: vault_pillar}
    return vault_pillar


def _get_paths(path_pattern, minion_id, pillar):
    """
    Get the paths that should be merged into the pillar dict
    """
    mappings = {"minion": minion_id, "pillar": pillar}

    paths = []
    try:
        for expanded_pattern in helpers.expand_pattern_lists(path_pattern, **mappings):
            paths.append(expanded_pattern.format(**mappings))
    except KeyError:
        log.warning("Could not resolve pillar path pattern %s", path_pattern)

    log.debug("%s vault pillar paths: %s", minion_id, paths)
    return paths



from textwrap import dedent

import pytest
import salt.utils.beacons
from saltfactories.utils import random_string

from tests.support.mysql import MySQLImage
from tests.support.mysql import create_mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_container  # pylint: disable=unused-import
from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_revoke
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def master_config_overrides():
    return {
        "vault": {
            "cache": {
                "backend": "disk",
            },
        },
    }


@pytest.fixture(scope="module")
def mysql_image():
    version = "10.3"
    return MySQLImage(
        name="mariadb",
        tag=version,
        container_id=random_string(f"mariadb-{version}-"),
    )


@pytest.fixture
def role_args_common():
    return {
        "db_name": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }


@pytest.fixture
def testrole():
    return {
        "default_ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def testreissuerole():
    return {
        "default_ttl": 180,
        "max_ttl": 180,
    }


@pytest.fixture
def teststaticrole(mysql_container):
    return {
        "db_name": "testdb",
        "rotation_period": 86400,
        "username": mysql_container.mysql_user,
    }


@pytest.fixture
def testdb(mysql_container, container_host_ref):
    return {
        "plugin_name": "mysql-database-plugin",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": "testrole,teststaticrole,testreissuerole",
        "username": "root",
        "password": mysql_container.mysql_passwd,
    }


@pytest.fixture(scope="module", autouse=True)
def db_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("database")
    yield
    assert vault_disable_secret_engine("database")


@pytest.fixture
def connection_setup(testdb):
    try:
        vault_write("database/config/testdb", **testdb)
        assert "testdb" in vault_list("database/config")
        yield
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")


@pytest.fixture(params=[["testrole"]], autouse=True)
def roles_setup(connection_setup, request, role_args_common):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            role_args.update(role_args_common)
            vault_write(f"database/roles/{role_name}", **role_args)
            assert role_name in vault_list("database/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("database/roles"):
                vault_delete(f"database/roles/{role_name}")
                assert role_name not in vault_list("database/roles")


@pytest.fixture(autouse=True)
def _cleanup():
    try:
        yield
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")
        if "testrole" in vault_list("database/roles"):
            vault_delete("database/roles/testrole")
            assert "testrole" not in vault_list("database/roles")
        if "teststaticrole" in vault_list("database/static-roles"):
            vault_delete("database/static-roles/teststaticrole")
            assert "teststaticrole" not in vault_list("database/static-roles")


@pytest.fixture
def _lease_beacon(master, salt_call_cli):
    state_contents = dedent("""
        Vault lease is cached:
          vault_db.creds_cached:
            - name: testrole
            # When the lease is requested, it is valid for
            # at least this amount of time
            - valid_for: 900
            - beacon: true
            - beacon_interval: 300
            - check_server: true
            # The beacon ensures this min_ttl, must be >= valid_for if set
            - min_ttl: 1200
            - meta: other.state.or.something.else
    """)
    states = master.state_tree.base.temp_file("vault_lease.sls", state_contents)
    try:
        with states:
            yield "vault_lease"
    finally:
        beacon = "vault_lease_db.database.dynamic.testrole.default"
        salt_call_cli.run("beacons.delete", beacon)
        res = salt_call_cli.run("beacons.list", return_yaml=False)
        assert res.returncode == 0
        assert beacon not in res.data


@pytest.fixture
def _lease_beacon_absent(master, salt_call_cli):
    state_contents = dedent("""
        Vault lease is not cached:
          vault_db.creds_uncached:
            - name: testrole
            - beacon: true
    """)
    states = master.state_tree.base.temp_file("vault_lease_absent.sls", state_contents)
    try:
        with states:
            yield "vault_lease_absent"
    finally:
        beacon = "vault_lease_db.database.dynamic.testrole.default"
        salt_call_cli.run("beacons.delete", beacon)
        res = salt_call_cli.run("beacons.list", return_yaml=False)
        assert res.returncode == 0
        assert beacon not in res.data


def test_creds_cached_mod_beacon(salt_call_cli, _lease_beacon, _lease_beacon_absent):
    """
    Ensure beacons can be added as part of caching a lease.
    """
    ret = salt_call_cli.run("state.apply", _lease_beacon)
    assert ret.returncode == 0

    # Ensure the lease is actually cached
    ckey = "db.database.dynamic.testrole.default"
    ret = salt_call_cli.run("vault_db.list_cached")
    assert ret.returncode == 0
    assert ret.data
    assert ckey in ret.data

    # Ensure the beacon has been created with the correct config
    # The beacon modules have a really weird API.
    # Dynamic beacons are added as an "opts" item.
    # We need to explicitly disable YAML output to get the Python object.
    # Each beacon config param is its own single-item (dict) list.
    # Also, most beacon functions are prone to crash on even the most basic
    # unexpected circumstances. Oh well. They are nice when they work.
    ret = salt_call_cli.run(
        "beacons.list", return_yaml=False, include_pillar=False, include_opts=True
    )
    assert ret.returncode == 0
    assert ret.data
    beacons_config = {}
    for beacon, conf in ret.data.items():
        beacons_config[beacon] = salt.utils.beacons.list_to_dict(conf)
    beacon = "vault_lease_db.database.dynamic.testrole.default"
    assert beacon in beacons_config
    assert beacons_config[beacon] == {
        "beacon_module": "vault_lease",
        "interval": 300,
        "leases": ckey,
        "min_ttl": 1200,
        "meta": "other.state.or.something.else",
        "check_server": True,
    }
    # I wanted to test if the beacon is executed with a high min_ttl
    # and low interval, but it seems beacons are not scheduled at all here.
    ret = salt_call_cli.run("state.apply", _lease_beacon_absent)
    assert ret.returncode == 0

    # Now remove the lease and its beacon again
    ret = salt_call_cli.run("vault_db.list_cached")
    assert ret.returncode == 0
    assert ckey not in ret.data

    ret = salt_call_cli.run(
        "beacons.list", return_yaml=False, include_pillar=False, include_opts=True
    )
    assert ret.returncode == 0
    assert beacon not in ret.data



"""
Runner functions supporting the Vault modules.

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import base64
import copy
import logging
import os
from collections.abc import Mapping
from pathlib import Path

import salt.cache
import salt.crypt
import salt.exceptions
import salt.pillar
import salt.utils.data
import salt.utils.json
import salt.utils.versions
from salt.defaults import NOT_SET
from salt.exceptions import SaltInvocationError
from salt.exceptions import SaltRunnerError
from salt.utils import immutabletypes

from saltext.vault.utils import vault
from saltext.vault.utils.vault import cache as vcache
from saltext.vault.utils.vault import factory
from saltext.vault.utils.vault import helpers
from saltext.vault.utils.vault.client import VaultClient
from saltext.vault.utils.versions import warn_until

log = logging.getLogger(__name__)

VALID_PARAMS = immutabletypes.freeze(
    {
        "approle": [
            "bind_secret_id",
            "secret_id_bound_cidrs",
            "secret_id_num_uses",
            "secret_id_ttl",
            "token_ttl",
            "token_max_ttl",
            "token_explicit_max_ttl",
            "token_num_uses",
            "token_no_default_policy",
            "token_period",
            "token_bound_cidrs",
        ],
        "token": [
            "ttl",
            "period",
            "explicit_max_ttl",
            "num_uses",
            "no_default_policy",
            "renewable",
        ],
    }
)

NO_OVERRIDE_PARAMS = immutabletypes.freeze(
    {
        "approle": [
            "bind_secret_id",
            "token_policies",
            "policies",
        ],
        "token": [
            "role_name",
            "policies",
            "meta",
        ],
    }
)


def auth_info():
    """
    .. versionadded:: 1.0.0

    Return information about the currently active Vault authentication.
    This includes token information and, if AppRole authentication is
    in use, information about the SecretID.

    CLI Example:

    .. code-block:: bash

        salt-run vault.auth_info
    """
    client = _get_master_client()
    info = {}
    info["token"] = client.get("auth/token/lookup-self")["data"]
    if _config("auth:method") == "approle":
        try:
            info["secret_id"] = _get_approle_api().read_secret_id(
                _config("auth:approle_name"),
                mount=_config("auth:approle_mount"),
                secret_id=str(client.auth.approle.secret_id),
            )
        except vault.VaultPermissionDeniedError:
            info["secret_id"] = (
                "Permission denied, allow API `create`, `update` access to "
                f"`auth/{_config('auth:approle_mount')}"
                f"/role/{_config('auth:approle_name')}/secret-id/lookup` "
                "to view info"
            )
    return info


def generate_token(
    minion_id,
    signature,
    impersonated_by_master=False,
    ttl=None,
    uses=None,
    upgrade_request=False,
):
    """
    .. deprecated:: 1.0.0

    Generate a Vault token for minion <minion_id>.

    minion_id
        The ID of the minion that requests a token.

    signature
        Cryptographic signature which validates that the request is indeed sent
        by the minion (or the master, see impersonated_by_master).

    impersonated_by_master
        If the master needs to create a token on behalf of the minion, this is
        True. This happens when the master generates minion pillars.

    ttl
        Ticket time to live in seconds, 1m minutes, or 2h hrs

    uses
        Number of times a token can be used

    upgrade_request
        In case the new runner endpoints have not been whitelisted for peer running,
        this endpoint serves as a gateway to :func:`vault.get_config <get_config>`.
        Defaults to False.
    """
    if upgrade_request:
        log.warning(
            "Detected minion fallback to old vault.generate_token peer run function. "
            "Please update your master peer_run configuration."
        )
        issue_params = {"explicit_max_ttl": ttl, "num_uses": uses}
        return get_config(minion_id, signature, impersonated_by_master, issue_params=issue_params)

    log.debug(
        "Token generation request for %s (impersonated by master: %s)",
        minion_id,
        impersonated_by_master,
    )
    _validate_signature(minion_id, signature, impersonated_by_master)
    try:
        warn_until(
            2,
            (
                "The vault.generate_token endpoint is deprecated and will be removed "
                "in version {version}. Please ensure your minions are running the "
                "Vault Salt extension as well."
            ),
        )

        if _config("issue:type") != "token":
            log.warning(
                "Master is not configured to issue tokens. Since the minion uses "
                "this deprecated endpoint, issuing token anyways."
            )

        issue_params = {}
        if ttl is not None:
            issue_params["explicit_max_ttl"] = ttl
        if uses is not None:
            issue_params["num_uses"] = uses

        token, _ = _generate_token(minion_id, issue_params=issue_params or None, wrap=False)
        ret = {
            "token": token["client_token"],
            "lease_duration": token["lease_duration"],
            "renewable": token["renewable"],
            "issued": token["creation_time"],
            "url": _config("server:url"),
            "verify": _config("server:verify"),
            "token_backend": _config("cache:backend"),
            "namespace": _config("server:namespace"),
        }
        if token["num_uses"] >= 0:
            ret["uses"] = token["num_uses"]

        return ret
    except Exception as err:  # pylint: disable=broad-except
        return {"error": f"{type(err).__name__}: {str(err)}"}


def generate_new_token(minion_id, signature, impersonated_by_master=False, issue_params=None):
    """
    .. versionadded:: 1.0.0

    Generate a Vault token for minion <minion_id>.

    minion_id
        The ID of the minion that requests a token.

    signature
        Cryptographic signature which validates that the request is indeed sent
        by the minion (or the master, see ``impersonated_by_master``).

    impersonated_by_master
        If the master needs to create a token on behalf of the minion, this is
        True. This happens when the master generates minion pillars.

    issue_params
        Dictionary of parameters for the generated tokens.
        See master configuration :vconf:`issue:token:params` for possible values.
        Requires :vconf:`issue:allow_minion_override_params` master configuration
        setting to be effective.
    """
    log.debug(
        "Token generation request for %s (impersonated by master: %s)",
        minion_id,
        impersonated_by_master,
    )
    _validate_signature(minion_id, signature, impersonated_by_master)
    try:
        if _config("issue:type") != "token":
            return {"expire_cache": True, "error": "Master does not issue tokens."}

        ret = {
            "server": _config("server"),
            "auth": {},
        }

        wrap = _config("issue:wrap")
        token, num_uses = _generate_token(minion_id, issue_params=issue_params, wrap=wrap)

        if wrap:
            ret.update(token)
            ret.update({"misc_data": {"num_uses": num_uses}})
        else:
            ret["auth"] = token

        return ret
    except Exception as err:  # pylint: disable=broad-except
        return {"error": f"{type(err).__name__}: {str(err)}"}


def _generate_token(minion_id, issue_params, wrap):
    endpoint = "auth/token/create"
    if _config("issue:token:role_name") is not None:
        endpoint += "/" + _config("issue:token:role_name")

    payload = _parse_issue_params(issue_params, issue_type="token")
    payload["policies"] = _get_policies_cached(
        minion_id,
        refresh_pillar=_config("policies:refresh_pillar"),
        expire=_config("policies:cache_time"),
    )

    if not payload["policies"]:
        raise SaltRunnerError("No policies matched minion.")

    payload["meta"] = _get_metadata(minion_id, _config("metadata:secret"))
    client = _get_master_client()
    log.trace("Sending token creation request to Vault.")
    res = client.post(endpoint, payload=payload, wrap=wrap)

    if wrap:
        return res.serialize_for_minion(), payload["num_uses"]
    if "num_uses" not in res["auth"]:
        # older vault versions do not include num_uses in output
        res["auth"]["num_uses"] = payload["num_uses"]
    token = vault.VaultToken(**res["auth"])
    return token.serialize_for_minion(), payload["num_uses"]


def get_config(
    minion_id,
    signature,
    impersonated_by_master=False,
    issue_params=None,
    config_only=False,
):
    """
    .. versionadded:: 1.0.0

    Return Vault configuration for minion <minion_id>.

    minion_id
        The ID of the minion that requests the configuration.

    signature
        Cryptographic signature which validates that the request is indeed sent
        by the minion (or the master, see impersonated_by_master).

    impersonated_by_master
        If the master needs to contact the Vault server on behalf of the minion, this is
        True. This happens when the master generates minion pillars.

    issue_params
        Parameters for credential issuance.
        Requires :vconf:`issue:allow_minion_override_params` master configuration
        setting to be effective.

    config_only
        In case the master is configured to issue tokens, do not include a new
        token in the response. This is used for configuration update checks.
        Defaults to false.
    """
    log.debug(
        "Config request for %s (impersonated by master: %s)",
        minion_id,
        impersonated_by_master,
    )
    _validate_signature(minion_id, signature, impersonated_by_master)
    try:
        minion_config = {
            "auth": {
                "method": _config("issue:type"),
                "token_lifecycle": _config("auth:token_lifecycle"),
            },
            "cache": _config("cache"),
            "client": _config("client"),
            "server": _config("server"),
            "wrap_info_nested": [],
        }
        wrap = _config("issue:wrap")

        if not config_only and _config("issue:type") == "token":
            minion_config["auth"]["token"], num_uses = _generate_token(
                minion_id,
                issue_params=issue_params,
                wrap=wrap,
            )
            if wrap:
                minion_config["wrap_info_nested"].append("auth:token")
                minion_config.update({"misc_data": {"token:num_uses": num_uses}})
        if _config("issue:type") == "approle":
            minion_config["auth"]["approle_mount"] = _config("issue:approle:mount")
            minion_config["auth"]["approle_name"] = minion_id
            minion_config["auth"]["secret_id"] = _config("issue:approle:params:bind_secret_id")
            minion_config["auth"]["role_id"] = _get_role_id(
                minion_id, issue_params=issue_params, wrap=wrap
            )
            if wrap:
                minion_config["wrap_info_nested"].append("auth:role_id")

        return minion_config
    except Exception as err:  # pylint: disable=broad-except
        return {"error": f"{type(err).__name__}: {str(err)}"}


def get_role_id(minion_id, signature, impersonated_by_master=False, issue_params=None):
    """
    .. versionadded:: 1.0.0

    Return the Vault role-id for minion <minion_id>. Requires the master to be configured
    to generate AppRoles for minions (:vconf:`issue:type`).

    minion_id
        The ID of the minion that requests a role-id.

    signature
        Cryptographic signature which validates that the request is indeed sent
        by the minion (or the master, see impersonated_by_master).

    impersonated_by_master
        If the master needs to create a token on behalf of the minion, this is
        True. This happens when the master generates minion pillars.

    issue_params
        Dictionary of configuration values for the generated AppRole.
        See master configuration :vconf:`issue:approle:params` for possible values.
        Requires :vconf:`issue:allow_minion_override_params` master configuration
        setting to be effective.
    """
    log.debug(
        "role-id request for %s (impersonated by master: %s)",
        minion_id,
        impersonated_by_master,
    )
    _validate_signature(minion_id, signature, impersonated_by_master)

    try:
        if _config("issue:type") != "approle":
            return {"expire_cache": True, "error": "Master does not issue AppRoles."}

        ret = {
            "server": _config("server"),
            "data": {},
        }

        wrap = _config("issue:wrap")
        role_id = _get_role_id(minion_id, issue_params=issue_params, wrap=wrap)
        if wrap:
            ret.update(role_id)
        else:
            ret["data"]["role_id"] = role_id
        return ret
    except Exception as err:  # pylint: disable=broad-except
        return {"error": f"{type(err).__name__}: {str(err)}"}


def _get_role_id(minion_id, issue_params, wrap):
    approle = _lookup_approle_cached(minion_id)
    issue_params_parsed = _parse_issue_params(issue_params)

    if approle is False or (
        helpers._get_salt_run_type(__opts__) != helpers.SALT_RUNTYPE_MASTER_IMPERSONATING
        and not _approle_params_match(approle, issue_params_parsed)
    ):
        # This means the role has to be created/updated first
        # create/update AppRole with role name <minion_id>
        # token_policies are set on the AppRole
        log.debug("Managing AppRole for %s.", minion_id)
        _manage_approle(minion_id, issue_params)
        # Make sure cached data is refreshed. Clearing the cache would suffice
        # here, but this branch should not be hit too often, so opt for simplicity.
        _lookup_approle_cached(minion_id, refresh=True)

    role_id = _lookup_role_id(minion_id, wrap=wrap)
    if role_id is False:
        raise SaltRunnerError(f"Failed to create AppRole for minion {minion_id}.")

    if approle is False:
        # This means the AppRole has just been created
        # create/update entity with name salt_minion_<minion_id>
        # metadata is set on the entity (to allow policy path templating)
        _manage_entity(minion_id)
        # ensure the new AppRole is mapped to the entity
        _manage_entity_alias(minion_id)

    if wrap:
        return role_id.serialize_for_minion()

    return role_id


def _approle_params_match(current, issue_params):
    """
    Check if minion-overridable AppRole parameters match
    """
    req = _parse_issue_params(issue_params)
    for var in set(VALID_PARAMS["approle"]) - set(NO_OVERRIDE_PARAMS["approle"]):
        if var in req and req[var] != current.get(var, NOT_SET):
            return False
    return True


def generate_secret_id(minion_id, signature, impersonated_by_master=False, issue_params=None):
    """
    .. versionadded:: 1.0.0

    Generate a Vault secret ID for minion <minion_id>. Requires the master to be
    configured to generate AppRoles for minions (:vconf:`issue:type`).

    minion_id
        The ID of the minion that requests a secret ID.

    signature
        Cryptographic signature which validates that the request is indeed sent
        by the minion (or the master, see ``impersonated_by_master``).

    impersonated_by_master
        If the master needs to create a token on behalf of the minion, this is
        True. This happens when the master generates minion pillars.

    issue_params
        Dictionary of configuration values for the generated AppRole.
        See master configuration :vconf:`issue:approle:params` for possible values.
        Requires :vconf:`issue:allow_minion_override_params` master configuration
        setting to be effective.
    """
    log.debug(
        "Secret ID generation request for %s (impersonated by master: %s)",
        minion_id,
        impersonated_by_master,
    )
    _validate_signature(minion_id, signature, impersonated_by_master)
    try:
        if _config("issue:type") != "approle":
            return {
                "expire_cache": True,
                "error": "Master does not issue AppRoles nor secret IDs.",
            }

        approle_meta = _lookup_approle_cached(minion_id)
        if approle_meta is False:
            raise vault.VaultNotFoundError(f"No AppRole found for minion {minion_id}.")

        if helpers._get_salt_run_type(
            __opts__
        ) != helpers.SALT_RUNTYPE_MASTER_IMPERSONATING and not _approle_params_match(
            approle_meta, issue_params
        ):
            _manage_approle(minion_id, issue_params)
            approle_meta = _lookup_approle_cached(minion_id, refresh=True)

        if not approle_meta["bind_secret_id"]:
            return {
                "expire_cache": True,
                "error": "Minion AppRole does not require a secret ID.",
            }

        ret = {
            "server": _config("server"),
            "data": {},
        }

        wrap = _config("issue:wrap")
        secret_id = _get_secret_id(minion_id, wrap=wrap)

        if wrap:
            ret.update(secret_id.serialize_for_minion())
        else:
            ret["data"] = secret_id.serialize_for_minion()

        ret["misc_data"] = {
            "secret_id_num_uses": approle_meta["secret_id_num_uses"],
        }
        return ret
    except vault.VaultNotFoundError as err:
        # when the role does not exist, make sure the minion requests
        # new configuration details to generate one
        return {
            "expire_cache": True,
            "error": f"{type(err).__name__}: {str(err)}",
        }
    except Exception as err:  # pylint: disable=broad-except
        return {"error": f"{type(err).__name__}: {str(err)}"}


def unseal():
    """
    Unseal the Vault server. Uses keys from the master config :vconf:`keys` .

    .. note::
        This function will send unseal keys until the API returns success.

    CLI Example:

    .. code-block:: bash

        salt-run vault.unseal
    """
    config = factory.parse_config(__opts__.get("vault", {}))
    client = VaultClient(**config["server"], **config["client"])

    for key in __opts__["vault"]["keys"]:
        ret = client.put("sys/unseal", payload={"key": key})
        # Return immediately after Vault is unsealed. No need to go over all the keys
        if ret["sealed"] is False:
            return True
    return False


def show_policies(minion_id, refresh_pillar=NOT_SET, expire=None):
    """
    Show the Vault policies that are applied to tokens for the given minion.

    minion_id
        The ID of the minion to show policies for.

    refresh_pillar
        Whether to refresh the pillar data when rendering templated policies.
        None will only refresh when the cached data is unavailable, boolean values
        force one behavior always.
        Defaults to :vconf:`policies:refresh_pillar` or None.

    expire
        Policy computation can be heavy in case pillar data is used in templated policies and
        it has not been cached. Therefore, a short-lived cache specifically for rendered policies
        is used. This specifies the expiration timeout in seconds.
        Defaults to :vconf:`policies:cache_time` or 60.

    .. note::

        When issuing AppRoles to minions, the shown policies are read from Vault
        configuration for the minion's AppRole and thus refresh_pillar/expire
        will not be honored.

    CLI Example:

    .. code-block:: bash

        salt-run vault.show_policies myminion
    """
    if _config("issue:type") == "approle":
        meta = _lookup_approle(minion_id)
        return meta["token_policies"]

    if refresh_pillar == NOT_SET:
        refresh_pillar = _config("policies:refresh_pillar")
    expire = expire if expire is not None else _config("policies:cache_time")
    return _get_policies_cached(minion_id, refresh_pillar=refresh_pillar, expire=expire)


def sync_approles(minions=None, up=False, down=False):
    """
    .. versionadded:: 1.0.0

    Sync minion AppRole parameters with current settings, including associated
    token policies.

    .. note::
        Only updates existing AppRoles. They are issued during the first request
        for one by the minion.
        Running this will reset minion overrides, which are reapplied automatically
        during the next request for authentication details.

    .. note::
        Unlike when issuing tokens, AppRole-associated policies are not regularly
        refreshed automatically. It is advised to schedule regular runs of this function.

    If no parameter is specified, will try to sync AppRoles for all known minions.

    CLI Example:

    .. code-block:: bash

        salt-run vault.sync_approles
        salt-run vault.sync_approles ecorp

    minions
        (List of) ID(s) of the minion(s) to update the AppRole for.
        Defaults to None.

    up
        Find all minions that are up and update their AppRoles.
        Defaults to False.

    down
        Find all minions that are down and update their AppRoles.
        Defaults to False.
    """
    if _config("issue:type") != "approle":
        raise SaltRunnerError("Master does not issue AppRoles to minions.")
    if minions is not None:
        if not isinstance(minions, list):
            minions = [minions]
    elif up or down:
        minions = []
        if up:
            minions.extend(__salt__["manage.list_state"]())
        if down:
            minions.extend(__salt__["manage.list_not_state"]())
    else:
        minions = _list_all_known_minions()

    for minion in set(minions) & set(list_approles()):
        _manage_approle(minion, issue_params=None)
        _lookup_approle_cached(minion, refresh=True)
        # Running multiple pillar renders in a loop would otherwise
        # falsely report a cyclic dependency (same loader context?)
        __opts__.pop("_vault_runner_is_compiling_pillar_templates", None)
    return True


def list_approles():
    """
    .. versionadded:: 1.0.0

    List all AppRoles that have been created by the Salt master.
    They are named after the minions.

    CLI Example:

    .. code-block:: bash

        salt-run vault.list_approles

    Required policy:

    .. code-block:: vaultpolicy

        path "auth/<mount>/role" {
            capabilities = ["list"]
        }
    """
    if _config("issue:type") != "approle":
        raise SaltRunnerError("Master does not issue AppRoles to minions.")
    api = _get_approle_api()
    return api.list_approles(mount=_config("issue:approle:mount"))


def sync_entities(minions=None, up=False, down=False):
    """
    .. versionadded:: 1.0.0

    Sync minion entities with current settings. Only updates entities for minions
    with existing AppRoles.

    .. note::
        This updates associated metadata only. Entities are created only
        when issuing AppRoles to minions (:vconf:`issue:type` == ``approle``).

    If no parameter is specified, will try to sync entities for all known minions.

    CLI Example:

    .. code-block:: bash

        salt-run vault.sync_entities

    minions
        (List of) ID(s) of the minion(s) to update the entity for.
        Defaults to None.

    up
        Find all minions that are up and update their associated entities.
        Defaults to False.

    down
        Find all minions that are down and update their associated entities.
        Defaults to False.
    """
    if _config("issue:type") != "approle":
        raise SaltRunnerError(
            "Master is not configured to issue AppRoles to minions, which is a "
            "requirement to use managed entities with Salt."
        )
    if minions is not None:
        if not isinstance(minions, list):
            minions = [minions]
    elif up or down:
        minions = []
        if up:
            minions.extend(__salt__["manage.list_state"]())
        if down:
            minions.extend(__salt__["manage.list_not_state"]())
    else:
        minions = _list_all_known_minions()

    for minion in set(minions) & set(list_approles()):
        _manage_entity(minion)
        # Running multiple pillar renders in a loop would otherwise
        # falsely report a cyclic dependency (same loader context?)
        __opts__.pop("_vault_runner_is_compiling_pillar_templates", None)
        entity = _lookup_entity_by_alias(minion)
        if not entity or entity["name"] != f"salt_minion_{minion}":
            log.info("Fixing association of minion AppRole to minion entity for %s.", minion)
            _manage_entity_alias(minion)
    return True


def list_entities():
    """
    .. versionadded:: 1.0.0

    List all entities that have been created by the Salt master.
    They are named ``salt_minion_{minion_id}``.

    CLI Example:

    .. code-block:: bash

        salt-run vault.list_entities

    Required policy:

    .. code-block:: vaultpolicy

        path "identity/entity/name" {
            capabilities = ["list"]
        }
    """
    if _config("issue:type") != "approle":
        raise SaltRunnerError("Master does not issue AppRoles to minions.")
    api = _get_identity_api()
    entities = api.list_entities()
    return [x for x in entities if x.startswith("salt_minion_")]


def show_entity(minion_id):
    """
    .. versionadded:: 1.0.0

    Show entity metadata for <minion_id>.

    CLI Example:

    .. code-block:: bash

        salt-run vault.show_entity db1
    """
    if _config("issue:type") != "approle":
        raise SaltRunnerError("Master does not issue AppRoles to minions.")
    api = _get_identity_api()
    return api.read_entity(f"salt_minion_{minion_id}")["metadata"]


def show_approle(minion_id):
    """
    .. versionadded:: 1.0.0

    Show AppRole configuration for <minion_id>.

    CLI Example:

    .. code-block:: bash

        salt-run vault.show_approle db1
    """
    if _config("issue:type") != "approle":
        raise SaltRunnerError("Master does not issue AppRoles to minions.")
    api = _get_approle_api()
    return api.read_approle(minion_id, mount=_config("issue:approle:mount"))


def cleanup_auth():
    """
    .. versionadded:: 1.0.0

    Removes AppRoles and entities associated with unknown minion IDs.
    Can only clean up entities if the AppRole still exists.

    .. warning::
        Make absolutely sure that the configured minion approle issue mount is
        exclusively dedicated to the Salt master, otherwise you might lose data
        by using this function! (:vconf:`issue:approle:mount`)

        This detects unknown existing AppRoles by listing all roles on the
        configured minion AppRole mount and deducting known minions from the
        returned list.

    CLI Example:

    .. code-block:: bash

        salt-run vault.cleanup_auth
    """
    ret = {"approles": [], "entities": []}

    for minion in set(list_approles()) - set(_list_all_known_minions()):
        if _fetch_entity_by_name(minion):
            _delete_entity(minion)
            ret["entities"].append(minion)
        _delete_approle(minion)
        ret["approles"].append(minion)
    return {"deleted": ret}


def clear_cache(master=True, minions=True):
    """
    .. versionadded:: 1.0.0

    Clears master cache of Vault-specific data. This can include:
    - AppRole metadata
    - rendered policies
    - cached authentication credentials for impersonated minions
    - cached KV metadata for impersonated minions

    CLI Example:

    .. code-block:: bash

        salt-run vault.clear_cache
        salt-run vault.clear_cache minions=false
        salt-run vault.clear_cache master=false minions='[minion1, minion2]'

    master
        Clear cached data for the master context.
        Includes cached master authentication data and KV metadata.
        Defaults to true.

    minions
        Clear cached data for minions on the master.
        Can include cached authentication credentials and KV metadata
        for pillar compilation as well as AppRole metadata and
        rendered policies for credential issuance.
        Defaults to true. Set this to a list of minion IDs to only clear
        cached data pertaining to thse minions.
    """
    config, _, _ = factory._get_connection_config("vault", __opts__, __context__, force_local=True)
    cache = vcache._get_cache_backend(config, __opts__)

    if cache is None:
        log.info("Vault cache clearance was requested, but no persistent cache is configured")
        return True

    if master:
        log.debug("Clearing master Vault cache")
        cache.flush("vault")
    if minions:
        for minion in cache.list("minions"):
            if minions is True or (isinstance(minions, list) and minion in minions):
                log.debug(f"Clearing master Vault cache for minion {minion}")
                cache.flush(f"minions/{minion}/vault")
    return True


def _config(key=None, default=vault.VaultException):
    ckey = "vault_master_config"
    if ckey not in __context__:
        __context__[ckey] = vault.parse_config(__opts__.get("vault", {}))

    if key is None:
        return __context__[ckey]
    val = salt.utils.data.traverse_dict(__context__[ckey], key, default)
    if val is vault.VaultException:
        raise vault.VaultException(f"Requested configuration value {key} does not exist.")
    return val


def _list_all_known_minions():
    return os.listdir(__opts__["pki_dir"] + "/minions")


def _validate_signature(minion_id, signature, impersonated_by_master):
    """
    Validate that either minion with id minion_id, or the master, signed the
    request
    """
    if not impersonated_by_master and _config("issue:block_minion_requests"):
        raise salt.exceptions.AuthenticationError(
            "Non-impersonated requests have been disabled in the config"
        )
    if not impersonated_by_master and __opts__.get("cluster_id") is not None:
        pki_dir = Path(__opts__["cluster_pki_dir"])
    else:
        pki_dir = Path(__opts__["pki_dir"])
    if impersonated_by_master:
        public_key = pki_dir / "master.pub"
    else:
        public_key = pki_dir / "minions" / minion_id

    log.trace("Validating signature for %s", minion_id)
    signature = base64.b64decode(signature)
    if not salt.crypt.verify_signature(str(public_key), minion_id, signature):
        raise salt.exceptions.AuthenticationError(
            f"Could not validate token request from {minion_id}"
        )
    log.trace("Signature ok")


# **kwargs because salt.cache.Cache does not pop "expire" from kwargs
def _get_policies(minion_id, refresh_pillar=None, **kwargs):  # pylint: disable=unused-argument
    """
    Get the policies that should be applied to a token for <minion_id>
    """
    grains, pillar = _get_minion_data(minion_id, refresh_pillar)
    mappings = {"minion": minion_id, "grains": grains, "pillar": pillar}

    policies = []
    for pattern in _config("policies:assign"):
        try:
            for expanded_pattern in helpers.expand_pattern_lists(pattern, **mappings):
                policies.append(expanded_pattern.format(**mappings).lower())  # Vault requirement
        except KeyError:
            log.warning("Could not resolve policy pattern %s for minion %s", pattern, minion_id)

    log.debug("%s policies: %s", minion_id, policies)
    return policies


def _get_policies_cached(minion_id, refresh_pillar=None, expire=60):
    # expiration of 0 disables cache
    if not expire:
        return _get_policies(minion_id, refresh_pillar=refresh_pillar)
    cbank = f"minions/{minion_id}/vault"
    ckey = "policies"
    cache = salt.cache.factory(__opts__)
    policies = cache.cache(
        cbank,
        ckey,
        _get_policies,
        expire=expire,
        minion_id=minion_id,
        refresh_pillar=refresh_pillar,
    )
    if not isinstance(policies, list):
        log.warning("Cached vault policies were not formed as a list. Refreshing.")
        cache.flush(cbank, ckey)
        policies = cache.cache(
            cbank,
            ckey,
            _get_policies,
            expire=expire,
            minion_id=minion_id,
            refresh_pillar=refresh_pillar,
        )
    return policies


def _get_minion_data(minion_id, refresh_pillar=None):
    _, grains, pillar = salt.utils.minions.get_minion_data(minion_id, __opts__)

    if grains is None:
        # In case no cached minion data is available, make sure the utils module
        # can distinguish a pillar refresh run impersonating a minion from running
        # on the master.
        grains = {"id": minion_id}
        # To properly refresh minion grains, something like this could be used:
        # __salt__["salt.execute"](minion_id, "saltutil.refresh_grains", refresh_pillar=False)
        # This is deliberately not done since grains should not be used to target
        # secrets anyways.

    # salt.utils.minions.get_minion_data only returns data from cache or None.
    # To make sure the correct policies are available, the pillar needs to be
    # refreshed. This can cause an infinite loop if the pillar data itself
    # depends on the vault execution module, which relies on this function.
    # By default, only refresh when necessary. Boolean values force one way.
    if refresh_pillar is True or (refresh_pillar is None and pillar is None):
        if __opts__.get("_vault_runner_is_compiling_pillar_templates"):
            raise SaltRunnerError(
                "Cyclic dependency detected while refreshing pillar for vault policy templating. "
                "This is caused by some pillar value relying on the vault execution module. "
                "Either remove the dependency from your pillar, disable refreshing pillar data "
                "for policy templating or do not use pillar values in policy templates."
            )
        local_opts = copy.deepcopy(__opts__)
        # Relying on opts for ext_pillars does not work properly (only the first one runs
        # correctly).
        extra_minion_data = {"_vault_runner_is_compiling_pillar_templates": True}
        local_opts.update(extra_minion_data)
        pillar = LazyPillar(local_opts, grains, minion_id, extra_minion_data=extra_minion_data)
    elif pillar is None:
        # Make sure pillar is a dict. Necessary because a check on LazyPillar would
        # refresh it unconditionally (even when no pillar values are used)
        pillar = {}

    return grains, pillar


def _get_metadata(minion_id, metadata_patterns, refresh_pillar=None):
    _, pillar = _get_minion_data(minion_id, refresh_pillar)
    mappings = {
        "minion": minion_id,
        "pillar": pillar,
        "jid": globals().get("__jid__", "<no jid set>"),
        "user": globals().get("__user__", "<no user set>"),
    }
    metadata = {}
    for key, pattern in metadata_patterns.items():
        metadata[key] = []
        composite_prefix = f"{key}__"

        # First, expand templates that make use of composite values (recursively)
        # until we have a list of patterns with simple values.
        try:
            expanded_patterns = helpers.expand_pattern_lists(pattern, **mappings)
        except (IndexError, KeyError):
            # If any template in the chain fails to render, skip all of them.
            # This was inherited from the old code and might be changed in a
            # new major release @FIXME
            log.warning(
                "Could not resolve metadata pattern %s for minion %s",
                pattern,
                minion_id,
            )
            expanded_patterns = []

        # Then substitute these simple template variables with their values.
        for expanded_pattern in expanded_patterns:
            try:
                metadata[key].append(expanded_pattern.format(**mappings))
            except (IndexError, KeyError):
                log.warning(
                    "Could not resolve metadata pattern %s for minion %s",
                    pattern,
                    minion_id,
                )

        # Since composite values are disallowed for metadata,
        # at least ensure the order of the comma-separated string
        # is predictable
        metadata[key].sort()
        # ... and assign the individual values to keys suffixed with their respective index.
        # A corresponding policy needs to account for this by repeating each assignment:
        #   path "salt/data/roles/{{identity.entity.metadata.role__0}}" {capabilities = ["read"]}
        #   path "salt/data/roles/{{identity.entity.metadata.role__1}}" {capabilities = ["read"]}
        #   ...
        if expanded_patterns == [pattern]:
            # Don't expand when the template did not involve a composite value.
            pass
        elif conflicting_keys := [
            mkey for mkey in metadata_patterns if mkey.startswith(composite_prefix)
        ]:
            # Don't risk overwriting explicitly set keys. It's unlikely, but check regardless.
            log.warning(
                "Skipping list expansion of entity metadata '%s' for minion '%s': "
                "Detected potentially conflicting key(s): %s'",
                key,
                minion_id,
                conflicting_keys,
            )
        else:
            for idx, value in enumerate(metadata[key]):
                metadata[f"{composite_prefix}{idx}"] = [value]

    log.debug("%s metadata: %s", minion_id, metadata)
    return {k: ",".join(v) for k, v in metadata.items()}


def _parse_issue_params(params, issue_type=None):
    if not _config("issue:allow_minion_override_params") or not isinstance(params, dict):
        params = {}

    # issue_type is used to override the configured type for minions using the old endpoint
    # TODO: remove this once the endpoint has been removed
    issue_type = issue_type or _config("issue:type")

    if issue_type not in VALID_PARAMS:
        raise SaltRunnerError("Invalid configuration for minion Vault authentication issuance.")

    configured_params = _config(f"issue:{issue_type}:params")
    ret = {}

    for valid_param in VALID_PARAMS[issue_type]:
        if valid_param in configured_params and configured_params[valid_param] is not None:
            ret[valid_param] = configured_params[valid_param]
        if (
            valid_param in params
            and valid_param not in NO_OVERRIDE_PARAMS[issue_type]
            and params[valid_param] is not None
        ):
            ret[valid_param] = params[valid_param]

    return ret


def _manage_approle(minion_id, issue_params):
    payload = _parse_issue_params(issue_params)
    # When the entity is managed during the same run, this can result in a duplicate
    # pillar refresh. Potential for optimization.
    payload["token_policies"] = _get_policies(minion_id, refresh_pillar=True)
    api = _get_approle_api()
    log.debug("Creating/updating AppRole for minion %s.", minion_id)
    return api.write_approle(minion_id, **payload, mount=_config("issue:approle:mount"))


def _delete_approle(minion_id):
    api = _get_approle_api()
    log.debug("Deleting approle for minion %s.", minion_id)
    return api.delete_approle(minion_id, mount=_config("issue:approle:mount"))


def _lookup_approle(minion_id, **kwargs):  # pylint: disable=unused-argument
    api = _get_approle_api()
    try:
        return api.read_approle(minion_id, mount=_config("issue:approle:mount"))
    except vault.VaultNotFoundError:
        return False


def _lookup_approle_cached(minion_id, expire=3600, refresh=False):
    # expiration of 0 disables cache
    if not expire:
        return _lookup_approle(minion_id)
    cbank = f"minions/{minion_id}/vault"
    ckey = "approle_meta"
    cache = salt.cache.factory(__opts__)
    if refresh:
        cache.flush(cbank, ckey)
    meta = cache.cache(
        cbank,
        ckey,
        _lookup_approle,
        expire=expire,
        minion_id=minion_id,
    )
    if not isinstance(meta, dict):
        log.warning(
            "Cached Vault AppRole meta information was not formed as a dictionary. Refreshing."
        )
        cache.flush(cbank, ckey)

        meta = cache.cache(
            cbank,
            ckey,
            _lookup_approle,
            expire=expire,
            minion_id=minion_id,
        )
    # Falsey values are always refreshed by salt.cache.Cache
    return meta


def _lookup_role_id(minion_id, wrap):
    api = _get_approle_api()
    try:
        return api.read_role_id(minion_id, mount=_config("issue:approle:mount"), wrap=wrap)
    except vault.VaultNotFoundError:
        return False


def _get_secret_id(minion_id, wrap):
    api = _get_approle_api()
    return api.generate_secret_id(
        minion_id,
        metadata=_get_metadata(minion_id, _config("metadata:secret")),
        mount=_config("issue:approle:mount"),
        wrap=wrap,
    )


def _lookup_entity_by_alias(minion_id):
    """
    This issues a lookup for the entity using the role-id and mount accessor,
    thus verifies that an entity and associated entity alias exists.
    """
    role_id = _lookup_role_id(minion_id, wrap=False)
    api = _get_identity_api()
    try:
        return api.read_entity_by_alias(alias=role_id, mount=_config("issue:approle:mount"))
    except vault.VaultNotFoundError:
        return False


def _fetch_entity_by_name(minion_id):
    api = _get_identity_api()
    try:
        return api.read_entity(name=f"salt_minion_{minion_id}")
    except vault.VaultNotFoundError:
        return False


def _manage_entity(minion_id):
    # When the approle is managed during the same run, this can result in a duplicate
    # pillar refresh. Potential for optimization.
    metadata = _get_metadata(minion_id, _config("metadata:entity"), refresh_pillar=True)
    api = _get_identity_api()
    return api.write_entity(f"salt_minion_{minion_id}", metadata=metadata)


def _delete_entity(minion_id):
    api = _get_identity_api()
    return api.delete_entity(f"salt_minion_{minion_id}")


def _manage_entity_alias(minion_id):
    role_id = _lookup_role_id(minion_id, wrap=False)
    api = _get_identity_api()
    log.debug("Creating entity alias for minion %s.", minion_id)
    try:
        return api.write_entity_alias(
            f"salt_minion_{minion_id}",
            alias_name=role_id,
            mount=_config("issue:approle:mount"),
        )
    except vault.VaultNotFoundError as err:
        raise SaltRunnerError(
            f"Cannot create alias for minion {minion_id}: no entity found."
        ) from err


def _get_approle_api():
    return vault.get_approle_api(__opts__, __context__, force_local=True)


def _get_identity_api():
    return vault.get_identity_api(__opts__, __context__, force_local=True)


def _get_master_client():
    # force_local is necessary when issuing credentials while impersonating
    # minions since the opts dict cannot be used to distinguish master from
    # minion in that case
    return vault.get_authd_client(__opts__, __context__, force_local=True)


def _revoke_token(token=None, accessor=None):
    if not token and not accessor:
        raise SaltInvocationError("Need either token or accessor to revoke token.")
    endpoint = "auth/token/revoke"
    if token:
        payload = {"token": token}
    else:
        endpoint += "-accessor"
        payload = {"accessor": accessor}
    client = _get_master_client()
    return client.post(endpoint, payload=payload)


class LazyPillar(Mapping):
    """
    Simulates a pillar dictionary. Only compiles the pillar
    once an item is requested.
    """

    def __init__(self, opts, grains, minion_id, extra_minion_data=None):
        self.opts = opts
        self.grains = grains
        self.minion_id = minion_id
        self.extra_minion_data = extra_minion_data or {}
        self._pillar = None

    def _load(self):
        log.info("Refreshing pillar for vault templating.")
        self._pillar = salt.pillar.get_pillar(
            self.opts,
            self.grains,
            self.minion_id,
            extra_minion_data=self.extra_minion_data,
        ).compile_pillar()

    def __getitem__(self, key):
        if self._pillar is None:
            self._load()
        return self._pillar[key]

    def __iter__(self):
        if self._pillar is None:
            self._load()
        yield from self._pillar

    def __len__(self):
        if self._pillar is None:
            self._load()
        return len(self._pillar)



import pytest
from saltfactories.utils import random_string

from saltext.vault.utils import vault
from tests.support.vault import vault_disable_auth_method
from tests.support.vault import vault_enable_auth_method

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        }
    }


@pytest.fixture
def fresh_auth_mount():
    name = random_string("fresh-mount", uppercase=False)
    vault_enable_auth_method("approle", [f"-path={name}"])
    try:
        yield name
    finally:
        vault_disable_auth_method(name)


@pytest.fixture
def approle_api(minion_opts):
    return vault.get_approle_api(minion_opts, {}, force_local=True)


def test_approle_api_list_empty_mount(approle_api, fresh_auth_mount):
    assert approle_api.list_approles(fresh_auth_mount) == []



from unittest.mock import Mock
from unittest.mock import patch

import pytest

from saltext.vault.utils import vault
from saltext.vault.utils.vault import auth as vauth
from saltext.vault.utils.vault import cache as vcache
from saltext.vault.utils.vault import client as vclient
from saltext.vault.utils.vault import leases as vleases


@pytest.fixture
def token(token_auth):
    return vleases.VaultToken(**token_auth["auth"])


@pytest.fixture
def token_invalid(token_auth):
    token_auth["auth"]["num_uses"] = 1
    token_auth["auth"]["use_count"] = 1
    return vleases.VaultToken(**token_auth["auth"])


@pytest.fixture
def token_unrenewable(token_auth):
    token_auth["auth"]["renewable"] = False
    return vleases.VaultToken(**token_auth["auth"])


@pytest.fixture
def secret_id(secret_id_response):
    return vleases.VaultSecretId(**secret_id_response["data"])


@pytest.fixture
def secret_id_invalid(secret_id_response):
    secret_id_response["data"]["secret_id_num_uses"] = 1
    secret_id_response["data"]["use_count"] = 1
    return vleases.VaultSecretId(**secret_id_response["data"])


@pytest.fixture(params=["secret_id"])
def approle(request):
    secret_id = request.param
    if secret_id is not None:
        secret_id = request.getfixturevalue(secret_id)
    return vauth.VaultAppRole("test-role-id", secret_id)


@pytest.fixture
def approle_invalid(secret_id_invalid):
    return vauth.VaultAppRole("test-role-id", secret_id_invalid)


@pytest.fixture
def token_store(token):
    store = Mock(spec=vauth.VaultTokenAuth)
    store.is_valid.return_value = True
    store.get_token.return_value = token
    return store


@pytest.fixture
def token_store_empty(token_store):
    token_store.is_valid.return_value = False
    token_store.get_token.side_effect = vault.VaultAuthExpired
    return token_store


@pytest.fixture
def token_store_empty_first(token_store, token):
    token_store.is_valid.side_effect = (False, True)
    token_store.get_token.side_effect = (token, vault.VaultException)
    return token_store


@pytest.fixture
def uncached():
    cache = Mock(spec=vcache.VaultAuthCache)
    cache.exists.return_value = False
    cache.get.return_value = None
    return cache


@pytest.fixture
def cached_token(uncached, token):
    uncached.exists.return_value = True
    uncached.get.return_value = token
    return uncached


@pytest.fixture
def client(token_auth):
    token_auth["auth"]["client_token"] = "new-test-token"
    client = Mock(spec=vclient.VaultClient)
    client.post.return_value = token_auth
    return client


def test_token_auth_uninitialized(uncached):
    """
    Test that an exception is raised when a token is requested
    and the authentication container was not passed a valid token.
    """
    auth = vauth.VaultTokenAuth(cache=uncached)
    uncached.get.assert_called_once()
    assert auth.is_valid() is False
    assert auth.is_renewable() is False
    auth.used()
    with pytest.raises(vault.VaultAuthExpired):
        auth.get_token()


def test_token_auth_cached(cached_token, token):
    """
    Test that tokens are read from cache.
    """
    auth = vauth.VaultTokenAuth(cache=cached_token)
    assert auth.is_valid()
    assert auth.get_token() == token


def test_token_auth_invalid_token(invalid_token):
    """
    Test that an exception is raised when a token is requested
    and the container's token is invalid.
    """
    auth = vauth.VaultTokenAuth(token=invalid_token)
    assert auth.is_valid() is False
    assert auth.is_renewable() is False
    with pytest.raises(vault.VaultAuthExpired):
        auth.get_token()


def test_token_auth_unrenewable_token(token_unrenewable):
    """
    Test that it is reported correctly by the container
    when a token is not renewable.
    """
    auth = vauth.VaultTokenAuth(token=token_unrenewable)
    assert auth.is_valid() is True
    assert auth.is_renewable() is False
    assert auth.get_token() == token_unrenewable


@pytest.mark.parametrize("num_uses", [0, 1, 10])
def test_token_auth_used_num_uses(uncached, token, num_uses):
    """
    Ensure that cache writes for use count are only done when
    num_uses is not 0 (= unlimited).
    Single-use tokens still require cache writes for updating
    ``uses``. The cache cannot be flushed here since
    exceptions might be used to indicate the token expiry
    to factory methods.
    """
    token = token.with_renewed(num_uses=num_uses)
    auth = vauth.VaultTokenAuth(cache=uncached, token=token)
    auth.used()
    if num_uses > 0:
        uncached.store.assert_called_once_with(token)
    else:
        uncached.store.assert_not_called()


@pytest.mark.usefixtures("time_stopped")
@pytest.mark.parametrize("num_uses", [0, 1, 10])
def test_token_auth_update_token(uncached, token, num_uses):
    """
    Ensure that partial updates to the token in use are possible
    and that the cache writes are independent from num_uses.
    Also ensure the token is treated as immutable
    """
    auth = vauth.VaultTokenAuth(cache=uncached, token=token)
    old_token = token
    old_token_ttl = old_token.duration
    auth.update_token({"num_uses": num_uses, "ttl": 8483})
    updated_token = token.with_renewed(num_uses=num_uses, ttl=8483)
    assert auth.token == updated_token
    assert old_token.duration == old_token_ttl
    uncached.store.assert_called_once_with(updated_token)


def test_token_auth_replace_token(uncached, token):
    """
    Ensure completely replacing the token is possible and
    results in a cache write. This is important when an
    InvalidVaultToken has to be replaced with a VaultToken,
    eg by a different authentication method.
    """
    auth = vauth.VaultTokenAuth(cache=uncached)
    assert isinstance(auth.token, vauth.InvalidVaultToken)
    auth.replace_token(token)
    assert isinstance(auth.token, vleases.VaultToken)
    assert auth.token == token
    uncached.store.assert_called_once_with(token)


@pytest.mark.parametrize("token", [False, True])
@pytest.mark.parametrize("approle", [False, True])
def test_approle_auth_is_valid(token, approle):
    """
    Test that is_valid reports true when either the token
    or the secret ID is valid
    """
    token = Mock(spec=vleases.VaultToken)
    token.is_valid.return_value = token
    approle = Mock(spec=vleases.VaultSecretId)
    approle.is_valid.return_value = approle
    auth = vauth.VaultAppRoleAuth(approle, None, token_store=token)
    assert auth.is_valid() is (token or approle)


def test_approle_auth_get_token_store_available(token_store, approle, token):
    """
    Ensure no login attempt is made when a cached token is available
    """
    auth = vauth.VaultAppRoleAuth(approle, None, token_store=token_store)
    with patch("saltext.vault.utils.vault.auth.VaultAppRoleAuth._login") as login:
        res = auth.get_token()
        login.assert_not_called()
        assert res == token


@pytest.mark.usefixtures("time_stopped")
def test_approle_auth_get_token_store_empty(token_store_empty, approle, token):
    """
    Ensure a token is returned if no cached token is available
    """
    auth = vauth.VaultAppRoleAuth(approle, None, token_store=token_store_empty)
    with patch("saltext.vault.utils.vault.auth.VaultAppRoleAuth._login") as login:
        login.return_value = token
        res = auth.get_token()
        login.assert_called_once()
        assert res == token


def test_approle_auth_get_token_invalid(token_store_empty, approle_invalid):
    """
    Ensure VaultAuthExpired is raised if a token request was made, but
    cannot be fulfilled
    """
    auth = vauth.VaultAppRoleAuth(approle_invalid, None, token_store=token_store_empty)
    with pytest.raises(vault.VaultAuthExpired):
        auth.get_token()


@pytest.mark.usefixtures("time_stopped")
@pytest.mark.parametrize("mount", ["approle", "salt_minions"])
@pytest.mark.parametrize("approle", ["secret_id", None], indirect=True)
def test_approle_auth_get_token_login(approle, mount, client, token_store_empty_first, token):
    """
    Ensure that login with secret-id returns a token that is passed to the
    token store/cache as well
    """
    auth = vauth.VaultAppRoleAuth(approle, client, mount=mount, token_store=token_store_empty_first)
    res = auth.get_token()
    assert res == token
    args, kwargs = client.post.call_args
    endpoint = args[0]
    payload = kwargs.get("payload", {})
    assert endpoint == f"auth/{mount}/login"
    assert "role_id" in payload
    if approle.secret_id is not None:
        assert "secret_id" in payload
    token_store_empty_first.replace_token.assert_called_once_with(res)


@pytest.mark.usefixtures("time_stopped")
@pytest.mark.parametrize("num_uses", [0, 1, 10])
def test_approle_auth_used_num_uses(
    token_store_empty_first, approle, client, uncached, num_uses, token
):
    """
    Ensure that cache writes for use count are only done when
    num_uses is not 0 (= unlimited)
    """
    approle.secret_id = approle.secret_id.with_renewed(num_uses=num_uses)
    auth = vauth.VaultAppRoleAuth(
        approle, client, cache=uncached, token_store=token_store_empty_first
    )
    res = auth.get_token()
    assert res == token
    if num_uses > 1:
        uncached.store.assert_called_once_with(approle.secret_id)
    elif num_uses:
        uncached.store.assert_not_called()
        uncached.flush.assert_called_once()
    else:
        uncached.store.assert_not_called()


@pytest.mark.usefixtures("time_stopped")
def test_approle_auth_used_locally_configured(
    token_store_empty_first, approle, client, uncached, token
):
    """
    Ensure that locally configured secret IDs are not cached.
    """
    approle.secret_id = vault.LocalVaultSecretId(**approle.secret_id.to_dict())
    auth = vauth.VaultAppRoleAuth(
        approle, client, cache=uncached, token_store=token_store_empty_first
    )
    res = auth.get_token()
    assert res == token
    uncached.store.assert_not_called()


def test_approle_allows_no_secret_id():
    """
    Ensure AppRole containers are still valid if no
    secret ID has been set (bind_secret_id can be set to False!)
    """
    role = vauth.VaultAppRole("test-role-id")
    assert role.is_valid()



"""
High-level utility functions for Vault (or OpenBao) interaction
"""

import logging

from saltext.vault.utils.vault.auth import InvalidVaultSecretId
from saltext.vault.utils.vault.auth import InvalidVaultToken
from saltext.vault.utils.vault.auth import LocalVaultSecretId
from saltext.vault.utils.vault.auth import VaultAppRole
from saltext.vault.utils.vault.exceptions import VaultAuthExpired
from saltext.vault.utils.vault.exceptions import VaultConfigExpired
from saltext.vault.utils.vault.exceptions import VaultException
from saltext.vault.utils.vault.exceptions import VaultInvocationError
from saltext.vault.utils.vault.exceptions import VaultNotFoundError
from saltext.vault.utils.vault.exceptions import VaultPermissionDeniedError
from saltext.vault.utils.vault.exceptions import VaultPreconditionFailedError
from saltext.vault.utils.vault.exceptions import VaultRateLimitExceededError
from saltext.vault.utils.vault.exceptions import VaultServerError
from saltext.vault.utils.vault.exceptions import VaultUnavailableError
from saltext.vault.utils.vault.exceptions import VaultUnsupportedOperationError
from saltext.vault.utils.vault.exceptions import VaultUnwrapException
from saltext.vault.utils.vault.factory import clear_cache
from saltext.vault.utils.vault.factory import get_approle_api
from saltext.vault.utils.vault.factory import get_authd_client
from saltext.vault.utils.vault.factory import get_identity_api
from saltext.vault.utils.vault.factory import get_kv
from saltext.vault.utils.vault.factory import get_lease_store
from saltext.vault.utils.vault.factory import parse_config
from saltext.vault.utils.vault.factory import update_config
from saltext.vault.utils.vault.leases import VaultLease
from saltext.vault.utils.vault.leases import VaultSecretId
from saltext.vault.utils.vault.leases import VaultToken
from saltext.vault.utils.vault.leases import VaultWrappedResponse

log = logging.getLogger(__name__)
logging.getLogger("requests").setLevel(logging.WARNING)


def query(
    method,
    endpoint,
    opts,
    context,
    payload=None,
    *,
    wrap=False,
    raise_error=True,
    safe_to_retry=None,
    is_unauthd=False,
    **kwargs,
):
    """
    Query the Vault API. Supplemental arguments to ``requests.request``
    can be passed as kwargs.

    method
        HTTP verb to use.

    endpoint
        API path to call (without leading ``/v1/``).

    opts
        Pass ``__opts__`` from the module.

    context
        Pass ``__context__`` from the module.

    payload
        Dictionary of payload values to send, if any.

    wrap
        Whether to request response wrapping. Should be a time string
        like ``30s`` or False (default).

    raise_error
        Whether to inspect the response code and raise exceptions.
        Defaults to True.

    safe_to_retry
        .. versionadded:: 1.4.0

        A boolean indicating whether this request is safe to retry (idempotent) or not.
        If not provided, defaults to guessing based on the HTTP method.
        Unsafe requests are not retried, unless :vconf:`client:retry_post` is enabled.

    is_unauthd
        Whether the queried endpoint is an unauthenticated one and hence
        does not deduct a token use. Only relevant for endpoints not found
        in ``sys``. Defaults to False.
    """
    client, config = get_authd_client(opts, context, get_config=True)
    try:
        return client.request(
            method,
            endpoint,
            payload=payload,
            wrap=wrap,
            raise_error=raise_error,
            safe_to_retry=safe_to_retry,
            is_unauthd=is_unauthd,
            **kwargs,
        )
    except VaultPermissionDeniedError:
        if not _check_clear(config, client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    client = get_authd_client(opts, context)
    return client.request(
        method,
        endpoint,
        payload=payload,
        wrap=wrap,
        raise_error=raise_error,
        safe_to_retry=safe_to_retry,
        is_unauthd=is_unauthd,
        **kwargs,
    )


def query_raw(
    method,
    endpoint,
    opts,
    context,
    payload=None,
    *,
    wrap=False,
    retry=True,
    is_unauthd=False,
    safe_to_retry=None,
    **kwargs,
):
    """
    Query the Vault API, returning the raw response object. Supplemental
    arguments to ``requests.request`` can be passed as kwargs.

    method
        HTTP verb to use.

    endpoint
        API path to call (without leading ``/v1/``).

    opts
        Pass ``__opts__`` from the module.

    context
        Pass ``__context__`` from the module.

    payload
        Dictionary of payload values to send, if any.

    retry
        Retry the query with cleared cache in case the permission
        was denied (to check for revoked cached credentials).
        Defaults to True.

        .. note::
            Affects handling of ``403 Forbidden`` responses by this function and
            is independent from client settings.

    wrap
        Whether to request response wrapping. Should be a time string
        like ``30s`` or False (default).

    safe_to_retry
        .. versionadded:: 1.4.0

        A boolean indicating whether this request is safe to retry (idempotent) or not.
        If not provided, defaults to guessing based on the HTTP method.
        Unsafe requests are not retried, unless :vconf:`client:retry_post` is enabled.

    is_unauthd
        Whether the queried endpoint is an unauthenticated one and hence
        does not deduct a token use. Only relevant for endpoints not found
        in ``sys``. Defaults to False.
    """
    client, config = get_authd_client(opts, context, get_config=True)
    res = client.request_raw(
        method,
        endpoint,
        payload=payload,
        wrap=wrap,
        safe_to_retry=safe_to_retry,
        is_unauthd=is_unauthd,
        **kwargs,
    )

    if not retry:
        return res

    if res.status_code == 403:
        if not _check_clear(config, client):
            return res

        # in case policies have changed
        clear_cache(opts, context)
        client = get_authd_client(opts, context)
        res = client.request_raw(
            method,
            endpoint,
            payload=payload,
            wrap=wrap,
            safe_to_retry=safe_to_retry,
            is_unauthd=is_unauthd,
            **kwargs,
        )
    return res


def is_v2(path, opts, context):
    """
    Determines if a given secret path is KV v1 or v2.
    """
    kv = get_kv(opts, context)
    return kv.is_v2(path)


def read_kv(path, opts, context, include_metadata=False, version=None):
    """
    Read secret at <path>.
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.read(path, include_metadata=include_metadata, version=version)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.read(path, include_metadata=include_metadata, version=version)


def read_kv_meta(path, opts, context):
    """
    Read secret metadata and version info at <path>.
    Requires KV v2.

    .. versionadded:: 1.2.0
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.read_meta(path)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.read_meta(path)


def write_kv(path, data, opts, context):
    """
    Write secret <data> to <path>.
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.write(path, data)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.write(path, data)


def patch_kv(path, data, opts, context):
    """
    Patch secret <data> at <path>.
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.patch(path, data)
    except VaultAuthExpired:
        # patching can consume several token uses when
        # 1) `patch` cap unvailable 2) KV v1 3) KV v2 w/ old Vault versions
        kv = get_kv(opts, context)
        return kv.patch(path, data)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.patch(path, data)


def delete_kv(path, opts, context, versions=None, all_versions=False):
    """
    Delete secret at <path>. For KV v2, versions can be specified,
    which will be soft-deleted.
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.delete(path, versions=versions, all_versions=all_versions)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.delete(path, versions=versions, all_versions=all_versions)


def restore_kv(path, opts, context, versions=None, all_versions=False):
    """
    Restore secret versions at <path>. Requires KV v2.
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.restore(path, versions=versions, all_versions=all_versions)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.restore(path, versions=versions, all_versions=all_versions)


def destroy_kv(path, versions, opts, context, all_versions=False):
    """
    Destroy secret <versions> at <path>. Requires KV v2.
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.destroy(path, versions, all_versions=all_versions)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.destroy(path, versions, all_versions=all_versions)


def wipe_kv(path, opts, context):
    """
    Completely remove all version history and data at <path>.
    Requires KV v2.

    .. versionadded:: 1.2.0
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.nuke(path)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.nuke(path)


def list_kv(path, opts, context):
    """
    List secrets at <path>.
    """
    kv, config = get_kv(opts, context, get_config=True)
    try:
        return kv.list(path)
    except VaultPermissionDeniedError:
        if not _check_clear(config, kv.client):
            raise

    # in case policies have changed
    clear_cache(opts, context)
    kv = get_kv(opts, context)
    return kv.list(path)


def _check_clear(config, client):
    """
    Called when encountering a VaultPermissionDeniedError.
    Decides whether caches should be cleared to retry with
    possibly updated token policies.
    """
    if config["cache"]["clear_on_unauthorized"]:
        return True
    try:
        # verify the current token is still valid
        return not client.token_valid(remote=True)
    except VaultAuthExpired:
        return True



import datetime
import json
import os
import platform
import shutil
import sys
import tempfile
from importlib import metadata
from pathlib import Path

import nox
from nox.command import CommandFailed
from nox.virtualenv import VirtualEnv

# Nox options
#  Reuse existing virtualenvs
nox.options.reuse_existing_virtualenvs = True
#  Don't fail on missing interpreters
nox.options.error_on_missing_interpreters = False
# Speed up all sessions by using uv if possible
if tuple(map(int, metadata.version("nox").split("."))) >= (2024, 3):
    nox.options.default_venv_backend = "uv|virtualenv"

# Python versions to test against
PYTHON_VERSIONS = ("3", "3.10")
# Be verbose when running under a CI context
CI_RUN = (
    os.environ.get("JENKINS_URL") or os.environ.get("CI") or os.environ.get("DRONE") is not None
)
PIP_INSTALL_SILENT = CI_RUN is False
SKIP_REQUIREMENTS_INSTALL = os.environ.get("SKIP_REQUIREMENTS_INSTALL", "0") == "1"
EXTRA_REQUIREMENTS_INSTALL = os.environ.get("EXTRA_REQUIREMENTS_INSTALL")

COVERAGE_REQUIREMENT = os.environ.get("COVERAGE_REQUIREMENT") or "coverage==7.13.2"
SALT_REQUIREMENT = os.environ.get("SALT_REQUIREMENT") or "salt>=3006"
if SALT_REQUIREMENT == "salt==master":
    SALT_REQUIREMENT = "git+https://github.com/saltstack/salt.git@master"

# Prevent Python from writing bytecode
os.environ["PYTHONDONTWRITEBYTECODE"] = "1"

# Global Path Definitions
REPO_ROOT = Path(__file__).resolve().parent
# Change current directory to REPO_ROOT
os.chdir(str(REPO_ROOT))

ARTIFACTS_DIR = REPO_ROOT / "artifacts"
# Make sure the artifacts directory exists
ARTIFACTS_DIR.mkdir(parents=True, exist_ok=True)
CUR_TIME = datetime.datetime.now().strftime("%Y%m%d%H%M%S.%f")
RUNTESTS_LOGFILE = ARTIFACTS_DIR / f"runtests-{CUR_TIME}.log"
COVERAGE_REPORT_DB = REPO_ROOT / ".coverage"
COVERAGE_REPORT_PROJECT = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "coverage-project.xml"
COVERAGE_REPORT_TESTS = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "coverage-tests.xml"
JUNIT_REPORT = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "junit-report.xml"


def _get_session_python_version_info(session):
    try:
        version_info = session._runner._real_python_version_info
    except AttributeError:
        session_py_version = session.run_always(
            "python",
            "-c",
            'import sys; sys.stdout.write("{}.{}.{}".format(*sys.version_info))',
            silent=True,
            log=False,
        )
        version_info = tuple(int(part) for part in session_py_version.split(".") if part.isdigit())
        session._runner._real_python_version_info = version_info
    return version_info


def _get_pydir(session):
    version_info = _get_session_python_version_info(session)
    if version_info < (3, 10):
        session.error("Only Python >= 3.10 is supported")
    return f"py{version_info[0]}.{version_info[1]}"


def _install_requirements(
    session,
    *passed_requirements,  # pylint: disable=unused-argument
    install_coverage_requirements=True,
    install_test_requirements=True,
    install_source=False,
    install_salt=True,
    install_extras=None,
):
    install_extras = install_extras or []
    no_progress = "--progress-bar=off"
    if isinstance(session._runner.venv, VirtualEnv) and session._runner.venv.venv_backend == "uv":
        no_progress = "--no-progress"
    if SKIP_REQUIREMENTS_INSTALL is False:
        # Always have the wheel package installed
        session.install(no_progress, "wheel", silent=PIP_INSTALL_SILENT)
        if install_coverage_requirements:
            session.install(no_progress, COVERAGE_REQUIREMENT, silent=PIP_INSTALL_SILENT)

        if install_salt:
            # Salt does not publish wheels and setuptools 75.6.0+ breaks requirements inclusion during builds,
            # so we need to constrain setuptools in the build environment. uv reads this from
            # pyproject.toml, but pip has no equivalent behavior.
            # We need delete=False for Windows. delete_on_close would work, but is Python 3.12+ only.
            with tempfile.NamedTemporaryFile(delete=False) as constraints_file:
                setuptools_constraint = "setuptools<75.6.0"
                constraints_file.write(setuptools_constraint.encode())
            env = {
                "PIP_CONSTRAINT": constraints_file.name,
            }
            try:
                session.install(no_progress, SALT_REQUIREMENT, silent=PIP_INSTALL_SILENT, env=env)
            finally:
                os.unlink(constraints_file.name)

        if install_test_requirements:
            install_extras.append("tests")

        if EXTRA_REQUIREMENTS_INSTALL:
            session.log(
                "Installing the following extra requirements because the "
                "EXTRA_REQUIREMENTS_INSTALL environment variable was set: "
                "EXTRA_REQUIREMENTS_INSTALL='%s'",
                EXTRA_REQUIREMENTS_INSTALL,
            )
            install_command = [no_progress]
            install_command += [req.strip() for req in EXTRA_REQUIREMENTS_INSTALL.split()]
            session.install(*install_command, silent=PIP_INSTALL_SILENT)

        if install_source:
            pkg = "."
            if install_extras:
                pkg += f"[{','.join(install_extras)}]"
            session.install("-e", pkg, silent=PIP_INSTALL_SILENT)
        elif install_extras:
            pkg = f".[{','.join(install_extras)}]"
            session.install(pkg, silent=PIP_INSTALL_SILENT)


@nox.session(python=PYTHON_VERSIONS)
def tests(session):
    _install_requirements(session, install_source=True)

    sitecustomize_dir = session.run("salt-factories", "--coverage", silent=True, log=False)
    python_path_env_var = os.environ.get("PYTHONPATH") or None
    if python_path_env_var is None:
        python_path_env_var = sitecustomize_dir
    else:
        python_path_entries = python_path_env_var.split(os.pathsep)
        if sitecustomize_dir in python_path_entries:
            python_path_entries.remove(sitecustomize_dir)
        python_path_entries.insert(0, sitecustomize_dir)
        python_path_env_var = os.pathsep.join(python_path_entries)

    env = {
        # The updated python path so that sitecustomize is importable
        "PYTHONPATH": python_path_env_var,
        # The full path to the .coverage data file. Makes sure we always write
        # them to the same directory
        "COVERAGE_FILE": str(COVERAGE_REPORT_DB),
        # Instruct sub processes to also run under coverage
        "COVERAGE_PROCESS_START": str(REPO_ROOT / ".coveragerc"),
    }

    # Support running functional/integration tests under Podman rootless
    if not CI_RUN and sys.platform.startswith("linux"):
        uid = os.getuid()
        if uid != 0:
            runtime_dir = Path(os.environ.get("XDG_RUNTIME_DIR", f"/run/user/{uid}"))
            podman_sock = runtime_dir / "podman" / "podman.sock"
            if podman_sock.exists():
                env["DOCKER_HOST"] = f"unix:{podman_sock}"
                env["CONTAINER_HOST_REF"] = "host.containers.internal"

    session.run("coverage", "erase")
    args = [
        "--rootdir",
        str(REPO_ROOT),
        f"--log-file={RUNTESTS_LOGFILE.relative_to(REPO_ROOT)}",
        "--log-file-level=debug",
        "--show-capture=no",
        f"--junitxml={JUNIT_REPORT}",
        "--showlocals",
        "-ra",
        "-s",
        "-vv",
    ]
    if session._runner.global_config.forcecolor:
        args.append("--color=yes")
    if not session.posargs:
        args.append("tests/")
    else:
        for arg in session.posargs:
            if arg.startswith("--color") and args[0].startswith("--color"):
                args.pop(0)
            args.append(arg)
        for arg in session.posargs:
            if arg.startswith("-"):
                continue
            if arg.startswith(f"tests{os.sep}"):
                break
            try:
                Path(arg).resolve().relative_to(REPO_ROOT / "tests")
                break
            except ValueError:
                continue
        else:
            args.append("tests/")
    try:
        session.run("coverage", "run", "-m", "pytest", *args, env=env)
    finally:
        # Always combine and generate the XML coverage report
        try:
            session.run("coverage", "combine")
        except CommandFailed:
            # Sometimes some of the coverage files are corrupt which would
            # trigger a CommandFailed exception
            pass
        # Generate report for salt code coverage
        session.run(
            "coverage",
            "xml",
            "-o",
            str(COVERAGE_REPORT_PROJECT),
            "--omit=tests/*",
            "--include=src/saltext/vault/*",
        )
        # Generate report for tests code coverage
        session.run(
            "coverage",
            "xml",
            "-o",
            str(COVERAGE_REPORT_TESTS),
            "--omit=src/saltext/vault/*",
            "--include=tests/*",
        )
        try:
            session.run("coverage", "report", "--show-missing", "--include=src/saltext/vault/*")
            # If you also want to display the code coverage report on the CLI
            # for the tests, comment the call above and uncomment the line below
            # session.run(
            #    "coverage", "report", "--show-missing",
            #    "--include=src/saltext/vault/*,tests/*"
            # )
        finally:
            # Move the coverage DB to artifacts/coverage in order for it to be archived by CI
            if COVERAGE_REPORT_DB.exists():
                shutil.move(str(COVERAGE_REPORT_DB), str(ARTIFACTS_DIR / COVERAGE_REPORT_DB.name))


class Tee:
    """
    Python class to mimic linux tee behaviour
    """

    def __init__(self, first, second):
        self._first = first
        self._second = second

    def write(self, buf):
        wrote = self._first.write(buf)
        self._first.flush()
        self._second.write(buf)
        self._second.flush()
        return wrote

    def fileno(self):
        return self._first.fileno()


def _lint(session, rcfile, flags, paths, tee_output=True):
    _install_requirements(
        session,
        install_salt=False,
        install_coverage_requirements=False,
        install_test_requirements=False,
        install_extras=["lint", "tests"],
    )

    if tee_output:
        session.run("pylint", "--version")
        pylint_report_path = os.environ.get("PYLINT_REPORT")

    cmd_args = ["pylint", f"--rcfile={rcfile}"] + list(flags) + list(paths)

    src_path = str(REPO_ROOT / "src")
    python_path_env_var = os.environ.get("PYTHONPATH") or None
    if python_path_env_var is None:
        python_path_env_var = src_path
    else:
        python_path_entries = python_path_env_var.split(os.pathsep)
        if src_path in python_path_entries:
            python_path_entries.remove(src_path)
        python_path_entries.insert(0, src_path)
        python_path_env_var = os.pathsep.join(python_path_entries)

    env = {
        # The updated python path so that the project is importable without installing it
        "PYTHONPATH": python_path_env_var,
        "PYTHONUNBUFFERED": "1",
    }

    cmd_kwargs = {"env": env}

    if tee_output:
        stdout = tempfile.TemporaryFile(mode="w+b")
        cmd_kwargs["stdout"] = Tee(stdout, sys.__stdout__)

    try:
        session.run(*cmd_args, **cmd_kwargs)
    finally:
        if tee_output:
            stdout.seek(0)
            contents = stdout.read()
            if contents:
                contents = contents.decode("utf-8")
                sys.stdout.write(contents)
                sys.stdout.flush()
                if pylint_report_path:
                    # Write report
                    with open(pylint_report_path, "w", encoding="utf-8") as wfh:
                        wfh.write(contents)
                    session.log("Report file written to %r", pylint_report_path)
            stdout.close()


def _lint_pre_commit(session, rcfile, flags, paths):
    if "VIRTUAL_ENV" not in os.environ:
        session.error(
            "This should be running from within a virtualenv and "
            "'VIRTUAL_ENV' was not found as an environment variable."
        )
    if "pre-commit" not in os.environ["VIRTUAL_ENV"]:
        session.error(
            "This should be running from within a pre-commit virtualenv and "
            f"'VIRTUAL_ENV'({os.environ['VIRTUAL_ENV']}) does not appear to be a pre-commit virtualenv."
        )

    # Let's patch nox to make it run inside the pre-commit virtualenv
    try:
        # nox >= 2024.03.02
        # pylint: disable=unexpected-keyword-arg
        venv = VirtualEnv(
            os.environ["VIRTUAL_ENV"],
            interpreter=session._runner.func.python,
            reuse_existing=True,
            venv_backend="venv",
        )
    except TypeError:
        # nox < 2024.03.02
        # pylint: disable=unexpected-keyword-arg
        venv = VirtualEnv(
            os.environ["VIRTUAL_ENV"],
            interpreter=session._runner.func.python,
            reuse_existing=True,
            venv=True,
        )
    session._runner.venv = venv
    _lint(session, rcfile, flags, paths, tee_output=False)


@nox.session(python="3")
def lint(session):
    """
    Run PyLint against the code and the test suite. Set PYLINT_REPORT to a path to capture output.
    """
    session.notify(f"lint-code-{session.python}")
    session.notify(f"lint-tests-{session.python}")


@nox.session(python="3", name="lint-code")
def lint_code(session):
    """
    Run PyLint against the code. Set PYLINT_REPORT to a path to capture output.
    """
    flags = ["--disable=I"]
    if session.posargs:
        paths = session.posargs
    else:
        paths = ["setup.py", "noxfile.py", "src/"]
    _lint(session, ".pylintrc", flags, paths)


@nox.session(python="3", name="lint-tests")
def lint_tests(session):
    """
    Run PyLint against the test suite. Set PYLINT_REPORT to a path to capture output.
    """
    flags = [
        "--disable=I,redefined-outer-name,no-member,missing-module-docstring,missing-function-docstring,missing-class-docstring,attribute-defined-outside-init,inconsistent-return-statements,too-few-public-methods,too-many-public-methods",
    ]
    if session.posargs:
        paths = session.posargs
    else:
        paths = ["tests/"]
    _lint(session, ".pylintrc", flags, paths)


@nox.session(python=False, name="lint-code-pre-commit")
def lint_code_pre_commit(session):
    """
    Run PyLint against the code. Set PYLINT_REPORT to a path to capture output.
    """
    flags = ["--disable=I"]
    if session.posargs:
        paths = session.posargs
    else:
        paths = ["setup.py", "noxfile.py", "src/"]
    _lint_pre_commit(session, ".pylintrc", flags, paths)


@nox.session(python=False, name="lint-tests-pre-commit")
def lint_tests_pre_commit(session):
    """
    Run PyLint against the code and the test suite. Set PYLINT_REPORT to a path to capture output.
    """
    flags = [
        "--disable=I,redefined-outer-name,no-member,missing-module-docstring,missing-function-docstring,missing-class-docstring,attribute-defined-outside-init,inconsistent-return-statements,too-few-public-methods,too-many-public-methods",
    ]
    if session.posargs:
        paths = session.posargs
    else:
        paths = ["tests/"]
    _lint_pre_commit(session, ".pylintrc", flags, paths)


def _get_docs_env(session):
    env = {}
    if not os.getenv("PYENCHANT_LIBRARY_PATH"):
        if sys.platform == "darwin" and platform.processor() == "arm":
            try:
                # Ensure docs build works on Apple Silicon, where the default
                # Homebrew lib path is not autodiscovered.
                # Needs `brew install enchant`.
                env["PYENCHANT_LIBRARY_PATH"] = str(
                    next(Path("/opt/homebrew/lib").glob("libenchant*.dylib"))
                )
            except StopIteration:
                session.warn(
                    "Failed to autodiscover enchant library. Ensure it's installed (e.g. brew install enchant). "
                    "If it's installed, set the PYENCHANT_LIBRARY_PATH environment variable to its dylib path."
                )
    return env


@nox.session(python="3")
def docs(session):
    """
    Build Docs
    """
    _install_requirements(
        session,
        install_coverage_requirements=False,
        install_test_requirements=False,
        install_source=True,
        install_extras=["docs"],
    )
    os.chdir("docs/")
    env = _get_docs_env(session)
    session.run("make", "clean", external=True)
    session.run("make", "linkcheck", "SPHINXOPTS=-W", external=True, env=env)
    session.run("make", "coverage", "SPHINXOPTS=-W", external=True, env=env)
    docs_coverage_file = os.path.join("_build", "html", "python.txt")
    if os.path.exists(docs_coverage_file):
        with open(docs_coverage_file) as rfh:  # pylint: disable=unspecified-encoding
            contents = rfh.readlines()[2:]
            if contents:
                session.error("\n" + "".join(contents))
    session.run("make", "html", "SPHINXOPTS=-W", external=True, env=env)
    os.chdir(str(REPO_ROOT))


@nox.session(name="docs-dev", python="3")
def docs_dev(session):
    """
    Build and serve the Sphinx HTML documentation, with live reloading on file changes, via sphinx-autobuild.

    Note: Only use this in INTERACTIVE DEVELOPMENT MODE. This SHOULD NOT be called
        in CI/CD pipelines, as it will hang.
    """
    _install_requirements(
        session,
        install_coverage_requirements=False,
        install_test_requirements=False,
        install_source=True,
        install_extras=["docs", "docsauto"],
    )

    build_dir = Path("docs", "_build", "html")

    # Allow specifying sphinx-autobuild options, like --host.
    args = ["--watch", "."] + session.posargs
    if not any(arg.startswith("--host") for arg in args):
        # If the user is overriding the host to something other than localhost,
        # it's likely they are rendering on a remote/headless system and don't
        # want the browser to open.
        args.append("--open-browser")
    args += ["docs", str(build_dir)]

    if build_dir.exists():
        shutil.rmtree(build_dir)

    env = _get_docs_env(session)
    session.run("sphinx-autobuild", *args, env=env)


@nox.session(name="docs-crosslink-info", python="3")
def docs_crosslink_info(session):
    """
    Report intersphinx cross links information
    """
    _install_requirements(
        session,
        install_coverage_requirements=False,
        install_test_requirements=False,
        install_source=True,
        install_extras=["docs"],
    )
    os.chdir("docs/")
    intersphinx_mapping = json.loads(
        session.run(
            "python",
            "-c",
            "import json; import conf; print(json.dumps(conf.intersphinx_mapping))",
            silent=True,
            log=False,
        )
    )
    intersphinx_mapping_list = ", ".join(list(intersphinx_mapping))
    try:
        mapping_entry = intersphinx_mapping[session.posargs[0]]
    except IndexError:
        session.error(
            f"You need to pass at least one argument whose value must be one of: {intersphinx_mapping_list}"
        )
    except KeyError:
        session.error(f"Only acceptable values for first argument are: {intersphinx_mapping_list}")
    session.run(
        "python", "-m", "sphinx.ext.intersphinx", mapping_entry[0].rstrip("/") + "/objects.inv"
    )
    os.chdir(str(REPO_ROOT))



"""
Vault PKI helpers

.. versionadded:: 1.1.0
"""

from datetime import datetime
from datetime import timedelta
from datetime import timezone

import salt.utils.x509 as x509util
from cryptography import x509 as cx509
from cryptography.hazmat.primitives import hashes
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.utils.vault.helpers import timestring_map


def check_cert_for_changes(
    current,
    issuer,
    private_key,
    common_name: str,
    encoding="pem",
    common_name_only=False,
    append_chain=None,
    private_key_passphrase=None,
    expire_tolerance=None,
    **kwargs,
) -> dict:

    changes = {}

    expire_tolerance = expire_tolerance or 0

    append_chain = append_chain or []
    if not isinstance(append_chain, list):
        append_chain = [append_chain]

    try:
        (
            current,
            current_encoding,
            current_chain,
            _,
        ) = x509util.load_cert(current, passphrase=None, get_encoding=True)
    except SaltInvocationError as err:
        if any(
            (
                "Could not deserialize binary data" in str(err),
                "Could not load PEM-encoded" in str(err),
            )
        ):
            changes["replaced"] = True
            return changes
        raise

    if encoding != current_encoding:
        changes["encoding"] = {
            "old": current_encoding,
            "new": encoding,
        }

    # Check common_name. This is always checked as a major
    # and required attribute for each certificate.
    current_cn = current.subject.get_attributes_for_oid(x509util.NAME_ATTRS_OID["CN"])[0].value
    if current_cn != common_name:
        changes.update({"subject": {"CN": {"old": current_cn, "new": common_name}}})

    # If we need to compare Common Name only we can skip this one
    if not common_name_only:
        for k, v in x509util.NAME_ATTRS_OID.items():
            # Just in case ignore CN attribute if passed by mistake
            if k == "CN":
                continue
            if k in kwargs:
                current_attr = current.subject.get_attributes_for_oid(v)
                if current_attr:
                    attr = current_attr[0]
                    if kwargs[k] != attr.value:
                        changes.update({"subject": {k: {"old": attr.value, "new": kwargs[k]}}})
                else:
                    changes.update({"subject": {k: {"old": "", "new": kwargs[k]}}})

    append_chain = [x509util.load_cert(x) for x in append_chain]
    # Filter self-signed CA, which shouldn't be in the chain.
    append_chain = [
        cert
        for cert in append_chain
        if cert.subject.rfc4514_string() != cert.issuer.rfc4514_string()
    ]

    if not compare_ca_chain(current_chain, append_chain):
        changes["ca_chain"] = True

    ca = x509util.load_cert(issuer)
    privkey = x509util.load_privkey(private_key, private_key_passphrase)

    changes.update(
        compare_cert_signing(
            current=current,
            signing_ca=ca,
            private_key=privkey,
        )
    )

    # Check if certificate should be renewed due to close to expiration
    try:
        curr_not_valid_after = current.not_valid_after_utc
    except AttributeError:
        curr_not_valid_after = current.not_valid_after.replace(tzinfo=timezone.utc)

    if curr_not_valid_after < datetime.now(timezone.utc) + timedelta(
        seconds=timestring_map(expire_tolerance, cast=int)
    ):
        changes["expiration"] = {
            "expire_in": (curr_not_valid_after - datetime.now(timezone.utc)).total_seconds(),
            "toleration": timestring_map(expire_tolerance, cast=int),
        }

    return changes


def compare_cert_signing(current: cx509.Certificate, signing_ca: cx509.Certificate, private_key):
    changes = {}

    if signing_ca and not x509util.verify_signature(current, signing_ca.public_key()):
        changes["signing_private_key"] = True

    # Check correctly if issuer is the same
    if _getattr_safe(signing_ca, "subject") != _getattr_safe(current, "issuer"):
        changes["issuer_name"] = {
            "old": _getattr_safe(current, "issuer").rfc4514_string(),
            "new": _getattr_safe(signing_ca, "subject").rfc4514_string(),
        }

    if not x509util.is_pair(current.public_key(), private_key):
        changes["private_key"] = True

    return changes


def compare_ca_chain(current: list[cx509.Certificate], new: list[cx509.Certificate]):
    if len(current) != len(new):
        return False
    for i, new_cert in enumerate(new):
        if new_cert.fingerprint(hashes.SHA256()) != current[i].fingerprint(hashes.SHA256()):
            return False
    return True


def dec2hex(decval):
    """
    Converts decimal values to nicely formatted hex strings
    """
    try:
        decval = int(decval)
    except (TypeError, ValueError) as exc:
        raise SaltInvocationError(f"input must be integer. got {type(decval)} instead") from exc

    if decval < 0:
        raise SaltInvocationError("input must be non-negative integer")

    return _pretty_hex(f"{decval:X}")


def _getattr_safe(obj, attr):
    try:
        return getattr(obj, attr)
    except AttributeError as err:
        # Since we cannot get the certificate object without signing,
        # we need to compare attributes marked as internal. At least
        # convert possible exceptions into some description.
        raise CommandExecutionError(
            f"Could not get attribute {attr} from {obj.__class__.__name__}. "
            "Did the internal API of cryptography change?"
        ) from err


def _pretty_hex(hex_str):
    """
    Nicely formats hex strings
    """
    if len(hex_str) % 2 != 0:
        hex_str = "0" + hex_str
    return ":".join([hex_str[i : i + 2] for i in range(0, len(hex_str), 2)]).upper()



import pytest


@pytest.fixture(scope="module")
def master(master):  # pragma: no cover
    with master.started():
        yield master


@pytest.fixture(scope="module")
def minion(minion):  # pragma: no cover
    with minion.started():
        yield minion


@pytest.fixture
def salt_run_cli(master):  # pragma: no cover
    return master.salt_run_cli()


@pytest.fixture
def salt_cli(master):  # pragma: no cover
    return master.salt_cli()


@pytest.fixture
def salt_call_cli(minion):  # pragma: no cover
    return minion.salt_call_cli()


@pytest.fixture(scope="module")
def salt_ssh_cli(
    master, salt_ssh_roster_file, sshd_config_dir, known_hosts_file
):  # pylint: disable=unused-argument; pragma: no cover
    return master.salt_ssh_cli(
        timeout=180,
        roster_file=salt_ssh_roster_file,
        target_host="localhost",
        client_key=str(sshd_config_dir / "client_key"),
    )



"""
Vault-specific cache classes
"""

import copy
import logging
import time
from abc import ABC
from abc import abstractmethod

import salt.cache

from saltext.vault.utils.vault import helpers
from saltext.vault.utils.vault import leases
from saltext.vault.utils.vault.exceptions import VaultConfigExpired
from saltext.vault.utils.vault.exceptions import VaultLeaseExpired

log = logging.getLogger(__name__)


def _get_config_cache(opts, context, cbank, ckey="config"):
    """
    Factory for VaultConfigCache to get around some
    chicken-and-egg problems
    """
    config = None
    if cbank in context and ckey in context[cbank]:
        config = context[cbank][ckey]
    else:
        cache = salt.cache.factory(opts)
        if cache.contains(cbank, ckey):
            # expiration check is done inside the class
            config = cache.fetch(cbank, ckey)
        elif opts.get("cache", "localfs") != "localfs":
            local_opts = copy.copy(opts)
            local_opts["cache"] = "localfs"
            cache = salt.cache.factory(local_opts)
            if cache.contains(cbank, ckey):
                # expiration check is done inside the class
                config = cache.fetch(cbank, ckey)

    return VaultConfigCache(
        context,
        cbank,
        ckey,
        opts,
        init_config=config,
        flush_exception=VaultConfigExpired,
    )


def _get_cache_backend(config, opts):
    if config["cache"]["backend"] == "session":
        return None
    if config["cache"]["backend"] in ("localfs", "disk", "file"):
        # cache.Cache does not allow setting the type of cache by param
        local_opts = copy.copy(opts)
        local_opts["cache"] = "localfs"
        return salt.cache.factory(local_opts)
    # this should usually resolve to localfs as well on minions,
    # but can be overridden by setting cache in the minion config
    return salt.cache.factory(opts)


def _get_cache_bank(opts, force_local=False, connection=True, session=False):
    minion_id = None
    # force_local is necessary because pillar compilation would otherwise
    # leak tokens between master and minions
    if not force_local and helpers._get_salt_run_type(opts) in (
        helpers.SALT_RUNTYPE_MASTER_IMPERSONATING,
        helpers.SALT_RUNTYPE_MASTER_PEER_RUN,
    ):
        minion_id = opts["grains"]["id"]
    prefix = "vault" if minion_id is None else f"minions/{minion_id}/vault"
    if session:
        return prefix + "/connection/session"
    if connection:
        return prefix + "/connection"
    return prefix


class CommonCache(ABC):
    """
    Base class that unifies context and other cache backends.
    """

    def __init__(self, context, cbank, cache_backend=None, ttl=None, flush_exception=None):
        self.context = context
        self.cbank = cbank
        self.cache = cache_backend
        self.ttl = ttl
        self.flush_exception = flush_exception

    @abstractmethod
    def flush(self):
        raise NotImplementedError()

    def _ckey_exists(self, ckey, flush=True):
        if self.cbank in self.context and ckey in self.context[self.cbank]:
            return True
        if self.cache is not None:
            if not self.cache.contains(self.cbank, ckey):
                return False
            if self.ttl is not None:
                updated = self.cache.updated(self.cbank, ckey)
                if int(time.time()) - updated >= self.ttl:
                    if flush:
                        log.debug(f"Cached data in {self.cbank}/{ckey} outdated, flushing.")
                        self.flush()
                    return False
            return True
        return False

    def _get_ckey(self, ckey, flush=True):
        if not self._ckey_exists(ckey, flush=flush):
            return None
        if self.cbank in self.context and ckey in self.context[self.cbank]:
            return self.context[self.cbank][ckey]
        if self.cache is not None:
            return self.cache.fetch(self.cbank, ckey) or None  # account for race conditions
        raise RuntimeError("This code path should not have been hit.")

    def _store_ckey(self, ckey, value):
        if self.cache is not None:
            self.cache.store(self.cbank, ckey, value)
        if self.cbank not in self.context:
            self.context[self.cbank] = {}
        self.context[self.cbank][ckey] = value

    def _flush(self, ckey=None):
        if not ckey and self.flush_exception is not None:
            # Flushing caches in Vault often requires an orchestrated effort
            # to ensure leases/sessions are terminated instead of left open.
            raise self.flush_exception()
        if self.cache is not None:
            self.cache.flush(self.cbank, ckey)
        if self.cbank in self.context:
            if ckey is None:
                self.context.pop(self.cbank)
            else:
                self.context[self.cbank].pop(ckey, None)
        # also remove sub-banks from context to mimic cache behavior
        if ckey is None:
            for bank in list(self.context):
                if bank.startswith(self.cbank):
                    self.context.pop(bank)

    def _list(self):
        ckeys = []
        if self.cbank in self.context:
            ckeys += list(self.context[self.cbank])
        if self.cache is not None:
            ckeys += self.cache.list(self.cbank)
        return set(ckeys)


class VaultCache(CommonCache):
    """
    Encapsulates session and other cache backends for a single domain
    like secret path metadata. Uses a single cache key.
    """

    def __init__(self, context, cbank, ckey, cache_backend=None, ttl=None, flush_exception=None):
        super().__init__(
            context,
            cbank,
            cache_backend=cache_backend,
            ttl=ttl,
            flush_exception=flush_exception,
        )
        self.ckey = ckey

    def exists(self, flush=True):
        """
        Check whether data for this domain exists
        """
        return self._ckey_exists(self.ckey, flush=flush)

    def get(self, flush=True):
        """
        Return the cached data for this domain or None
        """
        return self._get_ckey(self.ckey, flush=flush)

    def flush(self, cbank=False):
        """
        Flush the cache for this domain
        """
        return self._flush(self.ckey if not cbank else None)

    def store(self, value):
        """
        Store data for this domain
        """
        return self._store_ckey(self.ckey, value)


class VaultConfigCache(VaultCache):
    """
    Handles caching of received configuration
    """

    def __init__(
        self,
        context,
        cbank,
        ckey,
        opts,
        cache_backend_factory=_get_cache_backend,
        init_config=None,
        flush_exception=None,
    ):  # pylint: disable=super-init-not-called
        self.context = context
        self.cbank = cbank
        self.ckey = ckey
        self.opts = opts
        self.config = None
        self.cache = None
        self.ttl = None
        self.cache_backend_factory = cache_backend_factory
        self.flush_exception = flush_exception
        if init_config is not None:
            self._load(init_config)

    def exists(self, flush=True):
        """
        Check if a configuration has been loaded and cached
        """
        if self.config is None:
            return False
        return super().exists(flush=flush)

    def get(self, flush=True):
        """
        Return the current cached configuration
        """
        if self.config is None:
            return None
        return super().get(flush=flush)

    def flush(self, cbank=True):
        """
        Flush all connection-scoped data
        """
        if self.config is None:
            log.warning("Tried to flush uninitialized configuration cache. Skipping flush.")
            return
        # flush the whole connection-scoped cache by default
        super().flush(cbank=cbank)
        self.config = None
        self.cache = None
        self.ttl = None

    def _load(self, config):
        if self.config is not None:
            if (
                self.config["cache"]["backend"] != "session"
                and self.config["cache"]["backend"] != config["cache"]["backend"]
            ):
                self.flush()
        self.config = config
        self.cache = self.cache_backend_factory(self.config, self.opts)
        self.ttl = self.config["cache"]["config"]

    def store(self, value):
        """
        Reload cache configuration, then store the new Vault configuration,
        overwriting the existing one.
        """
        self._load(value)
        super().store(value)


class LeaseCacheMixin:
    """
    Mixin for auth and lease cache that checks validity
    and acts with hydrated objects
    """

    def __init__(self, *args, **kwargs):
        self.lease_cls = kwargs.pop("lease_cls", leases.VaultLease)
        self.expire_events = kwargs.pop("expire_events", None)
        super().__init__(*args, **kwargs)

    def _check_validity(self, lease_data, valid_for=0):
        lease = self.lease_cls(**lease_data)
        try:
            # is_valid on auth classes accounts for duration and uses
            if lease.is_valid(valid_for):
                log.debug("Using cached lease.")
                return lease
        except AttributeError:
            if lease.is_valid_for(valid_for):
                log.debug("Using cached lease.")
                return lease
        if self.expire_events is not None:
            raise VaultLeaseExpired(lease)
        return None


class VaultLeaseCache(LeaseCacheMixin, CommonCache):
    """
    Handles caching of Vault leases. Supports multiple cache keys.
    Checks whether cached leases are still valid before returning.
    Does not enforce for per-lease ``min_ttl``.
    """

    def get(self, ckey, valid_for=0, flush=True):
        """
        Returns valid cached lease data or None.
        Flushes cache if invalid by default.
        """
        data = self._get_ckey(ckey, flush=flush)
        if data is None:
            return data
        try:
            ret = self._check_validity(data, valid_for=valid_for)
        except VaultLeaseExpired as err:
            if self.expire_events is not None:
                self.expire_events(
                    tag=f"vault/lease/{ckey}/expire",
                    data={
                        "valid_for_less": (
                            valid_for if valid_for is not None else err.lease.min_ttl or 0
                        ),
                        "ttl_left": err.lease.ttl_left,
                        "meta": err.lease.meta,
                    },
                )
            ret = None
        if ret is None and flush:
            log.debug("Cached lease not valid anymore. Flushing cache.")
            self._flush(ckey)
        return ret

    def store(self, ckey, value):
        """
        Store a lease in cache
        """
        try:
            value = value.to_dict()
        except AttributeError:
            pass
        return self._store_ckey(ckey, value)

    def exists(self, ckey, flush=True):
        """
        Check whether a named lease exists in cache. Does not filter invalid ones,
        so fetching a reported one might still return None.
        """
        return self._ckey_exists(ckey, flush=flush)

    def flush(self, ckey=None):
        """
        Flush the lease cache or a single lease from the lease cache
        """
        return self._flush(ckey)

    def list(self):
        """
        List all cached leases. Does not filter invalid ones,
        so fetching a reported one might still return None.
        """
        return self._list()


class VaultAuthCache(LeaseCacheMixin, CommonCache):
    """
    Implements authentication secret-specific caches. Checks whether
    the cached secrets are still valid before returning.
    """

    def __init__(
        self,
        context,
        cbank,
        ckey,
        auth_cls,
        cache_backend=None,
        ttl=None,
        flush_exception=None,
    ):
        super().__init__(
            context,
            cbank,
            lease_cls=auth_cls,
            cache_backend=cache_backend,
            ttl=ttl,
            flush_exception=flush_exception,
        )
        self.ckey = ckey
        self.flush_exception = flush_exception

    def exists(self, flush=True):
        """
        Check whether data for this domain exists
        """
        return self._ckey_exists(self.ckey, flush=flush)

    def get(self, valid_for=0, flush=True):
        """
        Returns valid cached auth data or None.
        Flushes cache if invalid by default.
        """
        data = self._get_ckey(self.ckey, flush=flush)
        if data is None:
            return data
        ret = self._check_validity(data, valid_for=valid_for)
        if ret is None and flush:
            log.debug("Cached auth data not valid anymore. Flushing cache.")
            self.flush()
        return ret

    def store(self, value):
        """
        Store an auth credential in cache. Will overwrite possibly existing one.
        """
        try:
            value = value.to_dict()
        except AttributeError:
            pass
        return self._store_ckey(self.ckey, value)

    def flush(self, cbank=None):
        """
        Flush the cached auth credentials. If this is a token cache,
        flushing it will delete the whole session-scoped cache bank.
        """
        if self.lease_cls is leases.VaultToken:
            # flush the whole cbank (session-scope) if this is a token cache
            ckey = None
        else:
            ckey = None if cbank else self.ckey
        return self._flush(ckey)
